from django.db import models

from aw.config.main import config
from aw.config.hardcoded import SHORT_TIME_FORMAT
from aw.model.base import BaseModel, DEFAULT_NONE, CHOICES_BOOL, CHOICES_JOB_EXEC_STATUS
from aw.model.job_credential import JobGlobalCredentials
from aw.utils.util import get_choice_value_by_key, get_choice_key_by_value, datetime_from_db_str, is_null

CHOICES_REPOSITORY = [
    (1, 'Static'),
    (2, 'Git'),
]


class Repository(BaseModel):
    form_fields_git = [
        'name', 'git_origin', 'git_credentials', 'git_branch', 'git_isolate', 'git_lfs', 'git_limit_depth',
        'git_playbook_base',
        'git_hook_pre', 'git_hook_post', 'git_override_initialize', 'git_override_update',
    ]
    form_fields_static = ['name', 'static_path']
    form_fields = [
        'name', 'rtype', 'static_path', 'git_origin', 'git_credentials', 'git_branch', 'git_isolate', 'git_lfs',
        'git_limit_depth', 'git_hook_pre', 'git_hook_post', 'git_override_initialize', 'git_override_update',
        'git_playbook_base',
    ]
    api_fields_read = form_fields.copy()
    api_fields_read.extend([
        'id', 'rtype_name', 'time_update', 'status', 'status_name', 'log_stdout', 'log_stdout_url',
        'log_stderr', 'log_stderr_url',

    ])
    api_fields_write = form_fields

    name = models.CharField(max_length=100, null=False, blank=False)
    rtype = models.PositiveSmallIntegerField(choices=CHOICES_REPOSITORY)
    time_update = models.DateTimeField(**DEFAULT_NONE)
    status = models.PositiveSmallIntegerField(default=0, choices=CHOICES_JOB_EXEC_STATUS)
    log_stdout = models.CharField(max_length=300, **DEFAULT_NONE)
    log_stderr = models.CharField(max_length=300, **DEFAULT_NONE)

    static_path = models.CharField(max_length=500, **DEFAULT_NONE)

    git_origin = models.CharField(max_length=100, **DEFAULT_NONE)
    git_branch = models.CharField(max_length=100, **DEFAULT_NONE)
    git_isolate = models.BooleanField(choices=CHOICES_BOOL, default=False)
    git_lfs = models.BooleanField(choices=CHOICES_BOOL, default=False)
    git_limit_depth = models.PositiveIntegerField(**DEFAULT_NONE)
    git_hook_pre = models.CharField(max_length=1000, **DEFAULT_NONE)
    git_hook_post = models.CharField(max_length=1000, **DEFAULT_NONE)
    git_override_initialize = models.CharField(max_length=1000, **DEFAULT_NONE)
    git_override_update = models.CharField(max_length=1000, **DEFAULT_NONE)
    git_playbook_base = models.CharField(max_length=300, **DEFAULT_NONE)
    git_credentials = models.ForeignKey(
        JobGlobalCredentials, on_delete=models.SET_NULL, related_name='repo_fk_cred', null=True, blank=True,
    )

    @property
    def rtype_name(self) -> str:
        return self.rtype_name_from_id(self.rtype)

    @staticmethod
    def rtype_name_from_id(rtype) -> str:
        return get_choice_value_by_key(choices=CHOICES_REPOSITORY, find=rtype)

    @property
    def time_update_str(self) -> str:
        if is_null(self.time_update):
            return ''

        return datetime_from_db_str(dt=self.time_update, fmt=SHORT_TIME_FORMAT) + f" {config['timezone']}"

    @property
    def status_name(self) -> str:
        return self.status_name_from_id(self.status)

    @staticmethod
    def status_name_from_id(rtype) -> str:
        return get_choice_value_by_key(choices=CHOICES_JOB_EXEC_STATUS, find=rtype)

    @staticmethod
    def status_id_from_name(name: str) -> int:
        return get_choice_key_by_value(choices=CHOICES_JOB_EXEC_STATUS, find=name)

    @property
    def log_stdout_url(self) -> str:
        return f"/api/repository/log/{self.id}?type=stdout"

    @property
    def log_stderr_url(self) -> str:
        return f"/api/repository/log/{self.id}?type=stderr"

    def __str__(self) -> str:
        if self.rtype_name == 'Git':
            isolated = 'isolated ' if self.git_isolate else ''
            return f"{self.rtype_name.capitalize()} {isolated}repository - origin {self.git_origin}:{self.git_branch}"

        return f"{self.rtype_name.capitalize()} repository - path {self.static_path}"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['name'], name='repo_name_unique'),
        ]



# Generated by Django 5.0.2 on 2024-03-03 08:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aw", "0002_v0_0_13"),
    ]

    operations = [
        migrations.CreateModel(
            name="SchemaMetadata",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("schema_version", models.CharField(max_length=50)),
                (
                    "schema_version_prev",
                    models.CharField(
                        blank=True, default=None, max_length=50, null=True
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from aw.model.job import Job
from aw.model.job_credential import JobGlobalCredentials
from aw.base import USERS, GROUPS
from aw.model.repository import Repository


def choices_job() -> list[tuple]:
    # todo: only show jobs the user is privileged to view => get_viewable_jobs(user)
    return [(job.id, job.name) for job in Job.objects.all()]


def choices_global_credentials() -> list[tuple]:
    # todo: only show credentials the user is privileged to view => get_viewable_credentials(user)
    return [(credentials.id, credentials.name) for credentials in JobGlobalCredentials.objects.all()]


def choices_repositories() -> list[tuple]:
    # todo: only show credentials the user is privileged to view => get_viewable_credentials(user)
    return [(repo.id, repo.name) for repo in Repository.objects.all()]


def choices_user() -> list[tuple]:
    return [(user.id, user.username) for user in USERS.objects.all()]


def choices_group() -> list[tuple]:
    return [(group.id, group.name) for group in GROUPS.objects.all()]



from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers
from rest_framework.permissions import IsAuthenticated
from rest_framework_api_key.permissions import BaseHasAPIKey

from aw.model.api import AwAPIKey
from aw.base import USERS, GROUPS


class HasAwAPIKey(BaseHasAPIKey):
    model = AwAPIKey


API_PERMISSION = [IsAuthenticated | HasAwAPIKey]


# see: rest_framework_api_key.permissions.BaseHasAPIKey:get_from_header
def get_api_user(request) -> USERS:
    if isinstance(request.user, AnonymousUser):
        try:
            return AwAPIKey.objects.get_from_key(
                request.META.get(getattr(settings, 'API_KEY_CUSTOM_HEADER'))
            ).user

        except ObjectDoesNotExist:
            # invalid api key
            pass

    return request.user


class BaseResponse(serializers.Serializer):
    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass


class GenericResponse(BaseResponse):
    msg = serializers.CharField()


class GroupSerializer(serializers.ModelSerializer):
    class Meta:
        model = GROUPS


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = USERS


class LogDownloadResponse(BaseResponse):
    binary = serializers.CharField()



from pytz import all_timezones
from django import forms
from django.shortcuts import HttpResponse
from django.shortcuts import render
from django.contrib.auth.decorators import login_required

from aw.config.main import config
from aw.config.defaults import CONFIG_DEFAULTS
from aw.utils.http import ui_endpoint_wrapper
from aw.config.form_metadata import FORM_LABEL, FORM_HELP
from aw.config.environment import AW_ENV_VARS, AW_ENV_VARS_SECRET
from aw.model.system import SystemConfig
from aw.utils.deployment import deployment_dev


class SystemConfigForm(forms.ModelForm):
    class Meta:
        model = SystemConfig
        fields = SystemConfig.form_fields
        field_order = SystemConfig.form_fields
        labels = FORM_LABEL['system']['config']
        help_texts = FORM_HELP['system']['config']

    path_run = forms.CharField(max_length=500, initial=CONFIG_DEFAULTS['path_run'], required=True)
    path_play = forms.CharField(max_length=500, initial=CONFIG_DEFAULTS['path_play'], required=True)
    path_log = forms.CharField(max_length=500, initial=CONFIG_DEFAULTS['path_log'], required=True)
    path_ansible_config = forms.CharField(
        max_length=500, initial=CONFIG_DEFAULTS['path_ansible_config'], required=False,
    )
    path_ssh_known_hosts = forms.CharField(
        max_length=500, initial=CONFIG_DEFAULTS['path_ssh_known_hosts'], required=False,
    )
    timezone = forms.ChoiceField(
        required=False,
        widget=forms.Select,
        choices=[(tz, tz) for tz in sorted(all_timezones)],
        label=FORM_LABEL['system']['config']['timezone'],
    )
    debug = forms.BooleanField(initial=CONFIG_DEFAULTS['debug'] or deployment_dev())


@login_required
@ui_endpoint_wrapper
def system_config(request) -> HttpResponse:
    config_form = SystemConfigForm()
    form_method = 'put'
    form_api = 'config'

    config_form_html = config_form.render(
        template_name='forms/snippet.html',
        context={'form': config_form, 'existing': {
            key: config[key] for key in SystemConfig.form_fields
        }},
    )
    return render(
        request, status=200, template_name='system/config.html',
        context={
            'form': config_form_html, 'form_api': form_api, 'form_method': form_method,
            'env_vars': AW_ENV_VARS, 'env_labels': FORM_LABEL['system']['config'],
            'env_vars_secret': AW_ENV_VARS_SECRET,
            'env_vars_config': {key: config[key] for key in AW_ENV_VARS},
        }
    )



from platform import python_version
from datetime import datetime, timedelta
from time import time
from os import open as open_file
from pathlib import Path
from functools import lru_cache, wraps
from pkg_resources import get_distribution

from crontab import CronTab
from pytz import utc


from aw.config.main import config
from aw.config.hardcoded import SHORT_TIME_FORMAT
from aw.utils.util_no_config import set_timezone

# allow import from other modules
# pylint: disable=W0611
from aw.utils.util_no_config import is_set, is_null


def datetime_w_tz() -> datetime:
    return datetime.now(config.timezone)


def datetime_from_db(dt: (datetime, None)) -> (datetime, None):
    # datetime form db will always be UTC; convert it
    if not isinstance(dt, datetime):
        return None

    local_dt = dt.replace(tzinfo=utc).astimezone(config.timezone)
    return config.timezone.normalize(local_dt)


def datetime_from_db_str(dt: (datetime, None), fmt: str = SHORT_TIME_FORMAT) -> str:
    dt = datetime_from_db(dt)
    if not isinstance(dt, datetime):
        return ''

    return dt.strftime(fmt)


def get_next_cron_execution_sec(schedule: str) -> float:
    cron = CronTab(schedule)
    set_timezone(str(config.timezone))
    return cron.next(now=datetime_w_tz())


def get_next_cron_execution(schedule: str, wait_sec: (int, float) = None) -> datetime:
    if wait_sec is None:
        wait_sec = get_next_cron_execution_sec(schedule)

    return datetime.fromtimestamp(time() + wait_sec)


def get_next_cron_execution_str(schedule: str, wait_sec: (int, float) = None) -> str:
    return get_next_cron_execution(schedule, wait_sec).strftime(SHORT_TIME_FORMAT)


def _open_file_0600(path: (str, Path), flags):
    return open_file(path, flags, 0o600)


def write_file_0600(file: (str, Path), content: str):
    mode = 'w'
    if Path(file).is_file():
        mode = 'a'

    with open(file, mode, encoding='utf-8', opener=_open_file_0600) as _file:
        _file.write(content)


def _open_file_0640(path: (str, Path), flags):
    return open_file(path, flags, 0o640)


def write_file_0640(file: (str, Path), content: str):
    mode = 'w'
    if Path(file).is_file():
        mode = 'a'

    with open(file, mode, encoding='utf-8', opener=_open_file_0640) as _file:
        _file.write(content)


def get_ansible_versions() -> str:
    return (f"Python3: {python_version()} | "
            f"Ansible: {get_distribution('ansible').version} | "
            f"Ansible-Core: {get_distribution('ansible-core').version} | "
            f"Ansible-Runner: {get_distribution('ansible-runner').version} |"
            f"Ansible-WebUI: {config['version']}")


# source: https://realpython.com/lru-cache-python/
def timed_lru_cache(seconds: int, maxsize: int = 128):
    def wrapper_cache(func):
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = datetime.utcnow() + func.lifetime

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if datetime.utcnow() >= func.expiration:
                func.cache_clear()
                func.expiration = datetime.utcnow() + func.lifetime

            return func(*args, **kwargs)

        return wrapped_func

    return wrapper_cache


def get_choice_value_by_key(choices: list[tuple], find: any) -> (any, None):
    for k, v in choices:
        if k == find:
            return v

    return None


def get_choice_key_by_value(choices: list[tuple], find: any):
    for k, v in choices:
        if v == find:
            return k

    return None


def unset_or_null(data: dict, key: str) -> bool:
    return key not in data or is_null(data[key])



from django.shortcuts import render

from aw.utils.debug import log


class AnsibleConfigError(Exception):
    pass


class AnsibleRepositoryError(Exception):
    pass


def handler_log(request, msg: str, status: int):
    log(f"{request.build_absolute_uri()} - Got error {status} - {msg}")


def handler404(request, msg: str):
    handler_log(request=request, msg=msg, status=404)
    return render(request, 'error/404.html', context={'request': request, 'error_msg': msg})



# todo: some of these settings could be moved to the system-config later on

THREAD_JOIN_TIMEOUT = 3
INTERVAL_RELOAD = 10  # start/stop threads for configured jobs
INTERVAL_CHECK = 5  # check for queued jobs
LOGIN_PATH = '/a/login/'
LOGOUT_PATH = '/o/'
LOG_TIME_FORMAT = '%Y-%m-%d %H:%M:%S %z'
SHORT_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
FILE_TIME_FORMAT = '%Y-%m-%d_%H-%M-%S'
KEY_TIME_FORMAT = '%Y-%m-%d-%H-%M-%S'
MIN_SECRET_LEN = 30
JOB_EXECUTION_LIMIT = 20
GRP_MANAGER = {
    'job': 'AW Job Managers',
    'permission': 'AW Permission Managers',
    'repository': 'AW Repository Managers',
    'credentials': 'AW Credentials Managers',
    'system': 'AW System Managers',
}
REPO_CLONE_TIMEOUT = 300



from pathlib import Path
from shutil import rmtree
from os import symlink
from os import path as os_path
from os import remove as remove_file
from os import stat as os_stat

from ansible_runner import Runner, RunnerConfig
try:
    from ara.setup.callback_plugins import callback_plugins as ara_callback_plugins

except (ImportError, ModuleNotFoundError):
    ara_callback_plugins = None

from aw.config.main import config
from aw.utils.util import is_set, datetime_w_tz, write_file_0640
from aw.model.job import Job, JobExecution, JobExecutionResult, JobExecutionResultHost, JobError
from aw.model.job_credential import BaseJobCredentials
from aw.execute.util import update_status, overwrite_and_delete_file, decode_job_env_vars, \
    create_dirs, is_execution_status, config_error
from aw.utils.debug import log
from aw.execute.play_credentials import get_credentials_to_use, commandline_arguments_credentials, \
    write_pwd_file, get_pwd_file
from aw.execute.repository import ExecuteRepository

# see: https://ansible.readthedocs.io/projects/runner/en/latest/intro/


def _exec_log(execution: JobExecution, msg: str, level: int = 3):
    # todo: add execution logs to UI (?)
    log(
        msg=f"Job-execution '{execution.job}' @ {execution.result.time_start}: {msg}",
        level=level,
    )


def _commandline_arguments(job: Job, execution: JobExecution, path_run: Path) -> str:
    cmd_arguments = []
    if is_set(job.cmd_args):
        cmd_arguments.append(job.cmd_args)

    if is_set(execution.cmd_args):
        cmd_arguments.append(execution.cmd_args)

    if execution.mode_check or job.mode_check:
        cmd_arguments.append('--check')

    if execution.mode_diff or job.mode_diff:
        cmd_arguments.append('--diff')

    credentials = get_credentials_to_use(job=job, execution=execution)
    if credentials is not None:
        cmd_arguments.extend(
            commandline_arguments_credentials(credentials=credentials, path_run=path_run)
        )

    if is_set(config['path_ssh_known_hosts']) and \
            ' '.join(cmd_arguments).find('ansible_ssh_extra_args') == -1:
        if Path(config['path_ssh_known_hosts']).is_file():
            cmd_arguments.append(
                f"-e \"ansible_ssh_extra_args='-o UserKnownHostsFile={config['path_ssh_known_hosts']}'\""
            )

        else:
            _exec_log(execution=execution, msg='Ignoring known_hosts file because it does not exist', level=5)

    return ' '.join(cmd_arguments)


def _environmental_variables(job: Job, execution: JobExecution) -> dict:
    # merge global, job + execution env-vars
    env_vars = {}
    if is_set(config['ara_server']):
        if ara_callback_plugins is None:
            _exec_log(
                execution=execution,
                msg="Ignoring 'ara_server' setting because 'ara' module is not installed'",
                level=3,
            )

        else:
            env_vars['ANSIBLE_CALLBACK_PLUGINS'] = ara_callback_plugins
            env_vars['ARA_API_CLIENT'] = 'http'
            env_vars['ARA_API_SERVER'] = config['ara_server']

    if is_set(config['global_environment_vars']):
        env_vars = {
            **env_vars,
            **decode_job_env_vars(env_vars_csv=config['global_environment_vars'], src='Global')
        }

    if is_set(job.environment_vars.strip()):
        env_vars = {
            **env_vars,
            **decode_job_env_vars(env_vars_csv=job.environment_vars, src='Job')
        }

    if is_set(execution.environment_vars):
        env_vars = {
            **env_vars,
            **decode_job_env_vars(env_vars_csv=execution.environment_vars, src='Execution')
        }

    return env_vars


def _execution_or_job(job: Job, execution: JobExecution, attr: str):
    exec_val = getattr(execution, attr)
    if is_set(exec_val):
        return exec_val

    job_val = getattr(job, attr)
    if is_set(job_val):
        return job_val

    return None


def _runner_options(job: Job, execution: JobExecution, path_run: Path, project_dir: str) -> dict:
    verbosity = None
    if execution.verbosity != 0:
        verbosity = execution.verbosity

    elif job.verbosity != 0:
        verbosity = job.verbosity

    cmdline_args = _commandline_arguments(job=job, execution=execution, path_run=path_run)

    opts = {
        'project_dir': project_dir,
        'private_data_dir': path_run,
        'limit': _execution_or_job(job, execution, 'limit'),
        'tags': _execution_or_job(job, execution, 'tags'),
        'skip_tags': _execution_or_job(job, execution, 'tags_skip'),
        'verbosity': verbosity,
        'envvars': _environmental_variables(job=job, execution=execution),
        'cmdline': cmdline_args if is_set(cmdline_args) else None,
    }

    return opts


def runner_prep(job: Job, execution: JobExecution, path_run: Path, project_dir: str) -> dict:
    update_status(execution, status='Starting')

    opts = _runner_options(job=job, execution=execution, path_run=path_run, project_dir=project_dir)
    opts['playbook'] = job.playbook_file
    if is_set(job.inventory_file):
        opts['inventory'] = job.inventory_file.split(',')

    # https://docs.ansible.com/ansible/2.8/user_guide/playbooks_best_practices.html#directory-layout
    ppf = Path(opts['project_dir']) / opts['playbook']
    if not Path(ppf).is_file():
        config_error(f"Configured playbook not found: '{ppf}'")

    if 'inventory' in opts:
        for inventory in opts['inventory']:
            pi = Path(opts['project_dir']) / inventory
            if not Path(pi).exists():
                config_error(f"Configured inventory not found: '{pi}'")

    create_dirs(path=path_run, desc='run')
    create_dirs(path=config['path_log'], desc='log')

    credentials = get_credentials_to_use(job=job, execution=execution)
    for secret_attr in BaseJobCredentials.SECRET_ATTRS:
        write_pwd_file(credentials, attr=secret_attr, path_run=path_run)

    update_status(execution, status='Running')
    return opts


def runner_logs(cfg: RunnerConfig, log_files: dict):
    logs_src = {
        'stdout': os_path.join(cfg.artifact_dir, 'stdout'),
        'stderr': os_path.join(cfg.artifact_dir, 'stderr'),
    }

    for log_file in log_files.values():
        write_file_0640(file=log_file, content='')

    # link logs from artifacts to log-directory; have not found a working way of overriding the target files..
    for log_type in ['stdout', 'stderr']:
        try:
            symlink(log_files[log_type], logs_src[log_type])

        except FileExistsError:
            remove_file(logs_src[log_type])
            symlink(log_files[log_type], logs_src[log_type])


def runner_cleanup(execution: JobExecution, path_run: Path, exec_repo: ExecuteRepository):
    overwrite_and_delete_file(f"{path_run}/env/passwords")
    overwrite_and_delete_file(f"{path_run}/env/ssh_key")
    for attr in BaseJobCredentials.SECRET_ATTRS:
        overwrite_and_delete_file(get_pwd_file(path_run=path_run, attr=attr))

    try:
        exec_repo.cleanup_repository()

    except AttributeError as err:
        log(msg=f'Got error of repository cleanup: {err}')

    # clean empty log files
    for log_file in JobExecution.log_file_fields:
        log_file_path = getattr(execution, log_file)
        try:
            if os_stat(log_file_path).st_size == 0:
                remove_file(log_file_path)

        except (FileNotFoundError, TypeError):
            pass

    rmtree(path_run, ignore_errors=True)


def _run_stats(runner: Runner, result: JobExecutionResult) -> bool:
    any_task_failed = False
    for host in runner.stats['processed']:
        result_host = JobExecutionResultHost(hostname=host)

        result_host.unreachable = host in runner.stats['dark']
        result_host.tasks_skipped = runner.stats['skipped'][host] if host in runner.stats['skipped'] else 0
        result_host.tasks_ok = runner.stats['ok'][host] if host in runner.stats['ok'] else 0
        result_host.tasks_failed = runner.stats['failures'][host] if host in runner.stats['failures'] else 0
        result_host.tasks_ignored = runner.stats['ignored'][host] if host in runner.stats['ignored'] else 0
        result_host.tasks_rescued = runner.stats['rescued'][host] if host in runner.stats['rescued'] else 0
        result_host.tasks_changed = runner.stats['changed'][host] if host in runner.stats['changed'] else 0

        if result_host.unreachable:
            any_task_failed = True

        elif result_host.tasks_failed > 0:
            any_task_failed = True
            # todo: create errors

        result_host.result = result
        result_host.save()

    return any_task_failed


def parse_run_result(execution: JobExecution, result: JobExecutionResult, runner: Runner):
    result.time_fin = datetime_w_tz()
    result.failed = runner.errored
    result.save()

    any_task_failed = False
    if runner.stats is not None:
        any_task_failed = _run_stats(runner=runner, result=result)

    if runner.errored or runner.timed_out or runner.rc != 0 or any_task_failed:
        update_status(execution, status='Failed')

    else:
        status = 'Finished'
        if is_execution_status(execution, 'Stopping') or runner.canceled:
            status = 'Stopped'

        update_status(execution, status=status)


def failure(
        execution: JobExecution, exec_repo: ExecuteRepository, path_run: Path,
        result: JobExecutionResult, error_s: str, error_m: str
):
    update_status(execution, status='Failed')
    job_error = JobError(
        short=error_s,
        med=error_m,
    )
    job_error.save()
    result.time_fin = datetime_w_tz()
    result.failed = True
    result.error = job_error
    result.save()

    execution.save()
    runner_cleanup(execution=JobExecution ,path_run=path_run, exec_repo=exec_repo)



# Generated by Django 5.0.2 on 2024-02-28 18:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aw", "0001_v0_0_12"),
    ]

    operations = [
        migrations.AddField(
            model_name="systemconfig",
            name="ara_server",
            field=models.CharField(blank=True, default=None, max_length=300, null=True),
        ),
        migrations.AddField(
            model_name="systemconfig",
            name="global_environment_vars",
            field=models.CharField(
                blank=True, default=None, max_length=1000, null=True
            ),
        ),
        migrations.AlterField(
            model_name="job",
            name="credentials_needed",
            field=models.BooleanField(
                choices=[(True, "Yes"), (False, "No")], default=False
            ),
        ),
        migrations.AlterField(
            model_name="job",
            name="inventory_file",
            field=models.CharField(blank=True, default=None, max_length=300, null=True),
        ),
        migrations.AlterField(
            model_name="jobglobalcredentials",
            name="become_user",
            field=models.CharField(
                blank=True, default="root", max_length=100, null=True
            ),
        ),
        migrations.AlterField(
            model_name="jobusercredentials",
            name="become_user",
            field=models.CharField(
                blank=True, default="root", max_length=100, null=True
            ),
        ),
    ]



# pylint: disable=W0401,W0614,E0102
from django.db.backends.sqlite3.base import *


# NOTE: quick-patch for https://github.com/django/django/commit/a0204ac183ad6bca71707676d994d5888cf966aa
#   todo: remove once feature is available natively in django 5.1

class DatabaseWrapper(DatabaseWrapper):
    def _start_transaction_under_autocommit(self):
        self.cursor().execute('BEGIN IMMEDIATE')



from pathlib import Path
from os import environ
from sqlite3 import connect as db_connect
from sqlite3 import OperationalError

from yaml import safe_load as yaml_load
from yaml import YAMLError

try:
    from aw.config.main import config, VERSION

except ImportError:
    # pylint-django
    from aw.config.main import init_config
    init_config()
    from aw.config.main import config, VERSION


from aw.config.hardcoded import LOGIN_PATH
from aw.config.defaults import CONFIG_DEFAULTS
from aw.utils.deployment import deployment_dev, deployment_prod
from aw.config.environment import get_aw_env_var_or_default, auth_mode_saml
from aw.utils.debug import log

BASE_DIR = Path(__file__).resolve().parent.parent
TEMPLATE_DIRS = [
    BASE_DIR / 'aw' / 'templates/'
]

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'level': 'ERROR',
        },
    },
}

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
X_FRAME_OPTIONS = 'SAMEORIGIN'

INSTALLED_APPS = [
    'aw.apps.AwConfig',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # api
    'rest_framework',
    'rest_framework_api_key',
    'drf_spectacular',
    # styles
    'fontawesomefree',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django_auto_logout.middleware.auto_logout',
]

# Database
if deployment_prod():
    DB_FILE = Path(get_aw_env_var_or_default('db'))

    if DB_FILE.name.find('.') == -1 and not DB_FILE.exists():
        try:
            DB_FILE.mkdir(mode=0o750, parents=True, exist_ok=True)

        except (OSError, FileNotFoundError):
            raise ValueError(f"Unable to created database directory: '{DB_FILE}'")

    if DB_FILE.is_dir():
        DB_FILE = DB_FILE / 'aw.db'

else:
    dev_db_file = 'aw.dev.db' if deployment_dev() else 'aw.staging.db'
    if 'AW_DB' in environ:
        DB_FILE = Path(get_aw_env_var_or_default('db'))
        if DB_FILE.is_dir():
            DB_FILE = DB_FILE / dev_db_file

    else:
        DB_FILE = dev_db_file
        DB_FILE = BASE_DIR / DB_FILE

DATABASES = {
    'default': {
        'ENGINE': 'aw.db_sqlite_patched',
        # 'ENGINE': 'django.db.backends.sqlite3',  # todo: remove once feature is available natively in django 5.1
        'NAME': DB_FILE,
        'OPTIONS': {
            'timeout': 3,  # kill long-running write-requests fast; do not block whole application
            # 'transaction_mode': 'IMMEDIATE',  # waiting for django 5.1 :(
            # 'database is locked'; https://code.djangoproject.com/ticket/29280
        },
        'ATOMIC_REQUESTS': False,  # default
    }
}


def debug_mode() -> bool:
    # NOTE: only gets checked on startup
    if deployment_dev():
        return True

    if get_aw_env_var_or_default('debug'):
        return True

    try:
        if Path(DB_FILE).is_file():
            with db_connect(DB_FILE) as conn:
                return conn.execute('SELECT debug FROM aw_systemconfig').fetchall()[0][0] == 1

    except (IndexError, OperationalError):
        pass

    return False


DEBUG = debug_mode()

# WEB BASICS
PORT_WEB = get_aw_env_var_or_default('port')
LISTEN_ADDRESS = get_aw_env_var_or_default('address')
CSRF_TRUSTED_ORIGINS = [
    'http://localhost',
    f'http://localhost:{PORT_WEB}',
    'http://127.0.0.1',
    f'http://127.0.0.1:{PORT_WEB}',
]
if LISTEN_ADDRESS != '127.0.0.1':
    CSRF_TRUSTED_ORIGINS.extend([
        f'http://{LISTEN_ADDRESS}'
        f'http://{LISTEN_ADDRESS}:{PORT_WEB}'
        f'https://{LISTEN_ADDRESS}'
        f'https://{LISTEN_ADDRESS}:{PORT_WEB}'
    ])

if 'AW_PROXY' in environ:
    SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')
    USE_X_FORWARDED_HOST = True

ALLOWED_HOSTS = ['*']
if 'AW_HOSTNAMES' in environ:
    for hostname in environ['AW_HOSTNAMES'].split(','):
        ALLOWED_HOSTS.append(hostname)
        CSRF_TRUSTED_ORIGINS.extend([
            f'http://{hostname}',
            f'https://{hostname}',
            f'http://{hostname}:{PORT_WEB}',
            f'https://{hostname}:{PORT_WEB}',
        ])

CSRF_ALLOWED_ORIGINS = CSRF_TRUSTED_ORIGINS
CORS_ORIGINS_WHITELIST = CSRF_TRUSTED_ORIGINS

ROOT_URLCONF = 'aw.urls'
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
            'debug': DEBUG,
        },
    },
]
WSGI_APPLICATION = 'aw.main.app'

# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'},
    {'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'},
    {'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'},
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
        "OPTIONS": {
            "min_length": 10,
        },
    },
]

# Security
AUTO_LOGOUT = {
    'SESSION_TIME': config['session_timeout'],
}
SECRET_KEY = config['secret']

# Authentication
AUTH_MODE = CONFIG_DEFAULTS['auth_mode']
SAML2_AUTH = {}
if auth_mode_saml():
    INSTALLED_APPS.append('django_saml2_auth')

    with open(get_aw_env_var_or_default('saml_config_file'), 'r', encoding='utf-8') as _config:
        try:
            SAML2_AUTH = yaml_load(_config.read())

            # basic validation to help users find/fix their issues faster
            _ = SAML2_AUTH['ASSERTION_URL']
            _ = SAML2_AUTH['ENTITY_ID']
            _ = SAML2_AUTH['ATTRIBUTES_MAP']
            _ = SAML2_AUTH['ATTRIBUTES_MAP']['username']
            _ = SAML2_AUTH['ATTRIBUTES_MAP']['email']
            if 'METADATA_AUTO_CONF_URL' not in SAML2_AUTH and 'METADATA_LOCAL_FILE_PATH' not in SAML2_AUTH:
                raise KeyError('METADATA_AUTO_CONF_URL or METADATA_LOCAL_FILE_PATH')

            if ('TOKEN_REQUIRED' not in SAML2_AUTH or SAML2_AUTH['TOKEN_REQUIRED']) and \
                    'token' not in SAML2_AUTH['ATTRIBUTES_MAP']:
                raise KeyError('TOKEN_REQUIRED but not configured in ATTRIBUTES_MAP')

            if 'JWT_ALGORITHM' not in SAML2_AUTH:
                # for SSO-login page; internal communications
                SAML2_AUTH['JWT_ALGORITHM'] = CONFIG_DEFAULTS['jwt_algo']
                SAML2_AUTH['JWT_SECRET'] = CONFIG_DEFAULTS['jwt_secret']
                SAML2_AUTH['JWT_EXP'] = 60

        except YAMLError as err:
            log(msg=f"Failed to load SAML config: '{err}'", level=1)
            SAML2_AUTH = {}

        except KeyError as err:
            log(msg=f"Invalid SAML config: '{err}'", level=1)
            SAML2_AUTH = {}

    if len(SAML2_AUTH) > 0:
        AUTH_MODE = 'saml'

# Internationalization
LANGUAGE_CODE = 'en-us'
USE_I18N = True
USE_L10N = True
USE_TZ = True
TIME_ZONE = config['timezone']

# Static files (CSS, JavaScript, Images)
STATIC_URL = '/static/'
STATICFILES_DIRS = [BASE_DIR / 'aw' / 'static']
LOGIN_REDIRECT_URL = '/ui/jobs/manage'  # todo: change to '/ui' once dashboard is implemented
LOGOUT_REDIRECT_URL = LOGIN_PATH
handler403 = 'aw.utils.handlers.handler403'
handler500 = 'aw.utils.handlers.handler500'

# api
API_KEY_CUSTOM_HEADER = 'HTTP_X_API_KEY'
REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
    ),
    "DEFAULT_PERMISSION_CLASSES": [
        # 'rest_framework.permissions.AllowAny',
        'rest_framework.permissions.IsAuthenticated',
    ],
    'DEFAULT_SCHEMA_CLASS': 'drf_spectacular.openapi.AutoSchema',
}
SPECTACULAR_SETTINGS = {
    # 'TITLE': 'AW API',
    # 'DESCRIPTION': 'Your project description',
    'VERSION': VERSION,
    'SERVE_INCLUDE_SCHEMA': False,
    'SERVE_PERMISSIONS': ['rest_framework.permissions.IsAuthenticated'],
    'SWAGGER_UI_FAVICON_HREF': config['logo_url'],
    'APPEND_COMPONENTS': {
        'securitySchemes': {
            'apiKey': {
                'type': 'apiKey',
                'in': 'header',
                'name': 'X-Api-Key',
            },
            # 'session': {
            #     'type': 'apiKey',
            #     'in': 'cookie',
            #    'name': 'sessionid',
            # },
        },
    },
    'SECURITY': [
        {'apiKey': []},
        # {'session': []},
    ],
    'SWAGGER_UI_SETTINGS': {
        'displayOperationId': False,
    },
    'POSTPROCESSING_HOOKS': []
}

if deployment_dev():
    SPECTACULAR_SETTINGS['SWAGGER_UI_SETTINGS']['persistAuthorization'] = True



from django.apps import AppConfig
from django.dispatch import receiver
from django.db.backends.signals import connection_created


class AwConfig(AppConfig):
    name = 'aw'
    verbose_name = 'Ansible-WebUI'


# configuring sqlite at application startup/connection initialization
@receiver(connection_created)
def configure_sqlite(connection, **kwargs):
    if connection.vendor == 'sqlite':
        with connection.cursor() as cursor:
            # https://www.sqlite.org/pragma.html#pragma_journal_mode
            cursor.execute('PRAGMA journal_mode = WAL;')
            # https://www.sqlite.org/pragma.html#pragma_busy_timeout
            cursor.execute('PRAGMA busy_timeout = 5000;')
            # https://www.sqlite.org/pragma.html#pragma_synchronous
            cursor.execute('PRAGMA synchronous = NORMAL;')



# Generated by Django 5.0.2 on 2024-03-03 09:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aw", "0003_v0_0_14"),
    ]

    operations = [
        migrations.AlterField(
            model_name="systemconfig",
            name="timezone",
            field=models.CharField(default="UTC", max_length=300),
        ),
    ]



from os import environ
from importlib.metadata import version, PackageNotFoundError
from sys import stderr

from pytz import all_timezones, timezone, BaseTzInfo
from django.db.utils import IntegrityError, OperationalError
from django.core.exceptions import ImproperlyConfigured, AppRegistryNotReady

from aw.config.environment import get_aw_env_var, get_aw_env_var_or_default
from aw.utils.util_no_config import set_timezone, is_set
from aw.config.defaults import CONFIG_DEFAULTS


def __get_module_version() -> str:
    env_version = get_aw_env_var_or_default('version')
    if env_version is not None:
        return env_version

    try:
        return version('ansibleguy-webui')

    except PackageNotFoundError:
        # NOTE: not able to use aw.utils.debug.log_warn because of circular dependency
        stderr.write('\x1b[1;33mWARNING: Module version could not be determined!\x1b[0m\n')
        return '0.0.0'


VERSION = __get_module_version()


class Config:
    def __init__(self):
        set_timezone(self.get('timezone'))

    @staticmethod
    def _from_env_or_db(setting: str) -> any:
        env_var_value = get_aw_env_var(setting)
        if is_set(env_var_value):
            return env_var_value

        try:
            if 'AW_INIT' in environ and environ['AW_INIT'] == '1':
                # do not try to use ORM before django-init
                raise AppRegistryNotReady

            # pylint: disable=C0415
            from aw.model.system import get_config_from_db
            value = getattr(get_config_from_db(), str(setting))
            if value is not None:
                return value

        except (IntegrityError, OperationalError, ImproperlyConfigured, AppRegistryNotReady, ImportError,
                AttributeError):
            # if database not initialized or migrations missing; or env-only var
            pass

        if setting not in CONFIG_DEFAULTS:
            return None

        return CONFIG_DEFAULTS[setting]

    def get(self, setting: str) -> any:
        return self._from_env_or_db(setting)

    def __getitem__(self, setting):
        return self._from_env_or_db(setting)

    @property
    def timezone(self) -> BaseTzInfo:
        tz_str = self.get('timezone')

        if tz_str not in all_timezones:
            return timezone('GMT')

        return timezone(tz_str)

    def is_true(self, setting: str, fallback: bool = False) -> bool:
        val = self.get(setting)
        if val is None:
            return fallback

        if isinstance(val, bool):
            return val

        return str(val).lower() in ['1', 'true', 'y', 'yes']


def init_config():
    environ.setdefault('DJANGO_SETTINGS_MODULE', 'aw.settings')
    environ['PYTHONIOENCODING'] = 'utf8'
    environ['PYTHONUNBUFFERED'] = '1'
    environ['ANSIBLE_FORCE_COLOR'] = '1'

    # pylint: disable=W0601
    global config
    config = Config()



from aw.model.job import Job, JobQueue
from aw.utils.debug import log
from aw.base import USERS


def queue_get() -> (tuple[Job, USERS], None):
    next_queue_item = JobQueue.objects.order_by('-created').first()
    if next_queue_item is None:
        return None

    job, user = next_queue_item.job, next_queue_item.user
    next_queue_item.delete()
    return job, user


def queue_add(job: Job, user: USERS):
    log(msg=f"Job '{job.name}' added to execution queue", level=4)
    JobQueue(job=job, user=user).save()



import subprocess
from pathlib import Path
from os import environ
from functools import cache

from aw.settings import BASE_DIR
from aw.utils.debug import log
from aw.config.environment import AW_ENV_VARS_SECRET, AW_ENV_VARS


# pylint: disable=R0914
def process(
        cmd: (str, list), timeout_sec: int = None, shell: bool = False,
        cwd: Path = BASE_DIR, env: dict = None,
) -> dict:
    cmd_str = cmd
    if isinstance(cmd, list):
        cmd_str = ' '.join(cmd)

    log(msg=f"Executing command: '{cmd_str}'", level=6)

    # merge provided env with current env and hide secrets
    env_full = environ.copy()
    if env is not None:
        env_full = {**env_full, **env}

    for secret_var in AW_ENV_VARS_SECRET:
        for secret_env_var in AW_ENV_VARS[secret_var]:
            if secret_env_var in env_full:
                env_full.pop(secret_env_var)

    try:
        with subprocess.Popen(
            cmd,
            shell=shell,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cwd,
            env=env_full,
        ) as p:
            b_stdout, b_stderr = p.communicate(timeout=timeout_sec)
            stdout, stderr, rc = b_stdout.decode('utf-8').strip(), b_stderr.decode('utf-8').strip(), p.returncode

    except (subprocess.TimeoutExpired, subprocess.SubprocessError, subprocess.CalledProcessError,
            OSError, IOError) as error:
        stdout, stderr, rc = None, str(error), 1

    return {
        'stdout': stdout,
        'stderr': stderr,
        'rc': rc,
    }


@cache
def process_cache(
        cmd: str, timeout_sec: int = None, shell: bool = False,
        cwd: Path = BASE_DIR, env: dict = None,
) -> dict:
    # read-only commands which results can be cached
    return process(cmd=cmd.split(' '), timeout_sec=timeout_sec, shell=shell, cwd=cwd, env=env)



from django import forms
from django.shortcuts import HttpResponse
from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required

from aw.utils.http import ui_endpoint_wrapper_kwargs
from aw.model.permission import JobPermission, JobPermissionMapping, JobCredentialsPermissionMapping, \
    JobRepositoryPermissionMapping, JobPermissionMemberUser, JobPermissionMemberGroup
from aw.config.form_metadata import FORM_LABEL, FORM_HELP
from aw.views.base import choices_global_credentials, choices_job, choices_user, choices_group, choices_repositories


class SettingPermissionForm(forms.ModelForm):
    class Meta:
        model = JobPermission
        fields = JobPermission.form_fields
        field_order = JobPermission.form_fields
        labels = FORM_LABEL['settings']['permissions']
        help_texts = FORM_HELP['settings']['permissions']

    users = forms.MultipleChoiceField(
        required=False,
        widget=forms.SelectMultiple,
        choices=choices_user,
    )
    groups = forms.MultipleChoiceField(
        required=False,
        widget=forms.SelectMultiple,
        choices=choices_group,
    )
    jobs = forms.MultipleChoiceField(
        required=False,
        widget=forms.SelectMultiple,
        choices=choices_job,
    )
    credentials = forms.MultipleChoiceField(
        required=False,
        widget=forms.SelectMultiple,
        choices=choices_global_credentials,
    )
    repositories = forms.MultipleChoiceField(
        required=False,
        widget=forms.SelectMultiple,
        choices=choices_repositories,
    )


@login_required
@ui_endpoint_wrapper_kwargs
def setting_permission_edit(request, perm_id: int = None) -> HttpResponse:
    perm_form = SettingPermissionForm()
    form_method = 'post'
    form_api = 'permission'
    perm = {}

    if perm_id is not None and perm_id != 0:
        perm = JobPermission.objects.filter(id=perm_id).first()
        if perm is None:
            return redirect(f"/ui/settings/permissions?error=Permission with ID {perm_id} does not exist")

        data = perm.__dict__
        data['users'] = [link.user.id for link in JobPermissionMemberUser.objects.filter(permission=perm)]
        data['groups'] = [link.group.id for link in JobPermissionMemberGroup.objects.filter(permission=perm)]
        data['jobs'] = [link.job.id for link in JobPermissionMapping.objects.filter(permission=perm)]
        data['credentials'] = [
            link.credentials.id for link in JobCredentialsPermissionMapping.objects.filter(permission=perm)
        ]
        data['repositories'] = [
            link.repository.id for link in JobRepositoryPermissionMapping.objects.filter(permission=perm)
        ]
        perm = data
        form_method = 'put'
        form_api += f'/{perm_id}'

    perm_form_html = perm_form.render(
        template_name='forms/snippet.html',
        context={'form': perm_form, 'existing': perm},
    )
    return render(
        request, status=200, template_name='settings/permission_edit.html',
        context={'form': perm_form_html, 'form_api': form_api, 'form_method': form_method}
    )



from django.urls import path
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView

from aw.api_endpoints.key import APIKey, APIKeyItem
from aw.api_endpoints.job import APIJob, APIJobItem, APIJobExecutionItem, APIJobExecutionLogs, \
    APIJobExecutionLogFile, APIJobExecution
from aw.api_endpoints.permission import APIPermission, APIPermissionItem
from aw.api_endpoints.credentials import APIJobCredentials, APIJobCredentialsItem
from aw.api_endpoints.filesystem import APIFsBrowse, APIFsExists
from aw.api_endpoints.system import APISystemConfig
from aw.api_endpoints.repository import APIRepository, APIRepositoryItem, APIRepositoryLogFile

urlpatterns_api = [
    path('api/key/<str:token>', APIKeyItem.as_view()),
    path('api/key', APIKey.as_view()),
    path('api/job/<int:job_id>/<int:exec_id>/log/<int:line_start>', APIJobExecutionLogs.as_view()),
    path('api/job/<int:job_id>/<int:exec_id>/log', APIJobExecutionLogFile.as_view()),
    path('api/job/<int:job_id>/<int:exec_id>', APIJobExecutionItem.as_view()),
    path('api/job/<int:job_id>', APIJobItem.as_view()),
    path('api/job_exec', APIJobExecution.as_view()),
    path('api/job', APIJob.as_view()),
    path('api/permission/<int:perm_id>', APIPermissionItem.as_view()),
    path('api/permission', APIPermission.as_view()),
    path('api/credentials/<int:credentials_id>', APIJobCredentialsItem.as_view()),
    path('api/credentials', APIJobCredentials.as_view()),
    path('api/repository/log/<int:repo_id>', APIRepositoryLogFile.as_view()),
    path('api/repository/<int:repo_id>', APIRepositoryItem.as_view()),
    path('api/repository', APIRepository.as_view()),
    path('api/config', APISystemConfig.as_view()),
    path('api/fs/browse/<str:repository>', APIFsBrowse.as_view()),
    path('api/fs/exists', APIFsExists.as_view()),
    path('api/_schema/', SpectacularAPIView.as_view(), name='_schema'),
    path('api/_docs', SpectacularSwaggerView.as_view(url_name='_schema'), name='swagger-ui'),
]



from typing import Callable

from django.shortcuts import HttpResponse, redirect

from aw.config.hardcoded import LOGIN_PATH


def deny_request(request) -> (bool, HttpResponse):
    if request.method not in ['GET', 'POST', 'PUT']:
        return True, HttpResponse(status=405)

    return False, None


def ui_endpoint_wrapper_auth(func) -> Callable:
    def wrapper(request, *args, **kwargs):
        del args
        del kwargs

        bad, deny = deny_request(request)
        if bad:
            return deny

        return func(request)

    return wrapper


def ui_endpoint_wrapper(func) -> Callable:
    def wrapper(request, *args, **kwargs):
        del args
        del kwargs

        bad, deny = deny_request(request)
        if bad:
            return deny

        if not request.user.is_authenticated:
            return redirect(LOGIN_PATH)

        return func(request)

    return wrapper


def ui_endpoint_wrapper_kwargs(func) -> Callable:
    def wrapper(request, *args, **kwargs):
        del args

        bad, deny = deny_request(request)
        if bad:
            return deny

        if not request.user.is_authenticated:
            return redirect(LOGIN_PATH)

        return func(request, **kwargs)

    return wrapper



from aw.config.hardcoded import LOGOUT_PATH

NAVIGATION = {
    'left': {
        # 'Dashboard': '/ui/',
        'Jobs': {
            'Manage': '/ui/jobs/manage',
            'Logs': '/ui/jobs/log',
            'Credentials': '/ui/jobs/credentials',
            'Repositories': '/ui/jobs/repository',
        },
        'Settings': {
            'API Keys': '/ui/settings/api_keys',
            'Permissions': '/ui/settings/permissions',
            # 'Ansible': '/ui/settings/ansible',
        },
        'System': {
            'Admin': '/ui/system/admin/',
            'API Docs': '/ui/system/api_docs',
            'Config': '/ui/system/config',
            'Environment': '/ui/system/environment',
            'Password': '/a/password_change/',
        },
    },
    'right': {
        'GH': {
            'element': '<i class="fab fa-github-square fa-2x aw-nav-right-icon" title="GitHub"></i>',
            'url': 'https://github.com/ansibleguy/webui',
            'login': False,
        },
        'DON': {
            'element': '<i class="fas fa-coins fa-2x aw-nav-right-icon" title="Sponsor"></i>',
            'url': 'https://github.com/sponsors/ansibleguy',
            'login': False,
        },
        'BUG': {
            'element': '<i class="fas fa-bug fa-2x aw-nav-right-icon" title="Report bug"></i>',
            'url': 'https://github.com/ansibleguy/webui/issues',
            'login': False,
        },
        'DOC': {
            'element': '<i class="fas fa-book fa-2x aw-nav-right-icon" title="Documentation"></i>',
            'url': 'https://webui.ansibleguy.net/',
            'login': False,
        },
        'LO': {
            'element': '<i class="fas fa-sign-out-alt fa-2x aw-nav-right-icon aw-nav-right-icon-logout" '
                       'title="Logout"></i>',
            'url': LOGOUT_PATH,
            'method': 'post',
            'login': True,
        },
    }
}



from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import IntegrityError
from rest_framework.views import APIView
from rest_framework import serializers
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema, OpenApiResponse, OpenApiParameter

from aw.config.hardcoded import JOB_EXECUTION_LIMIT
from aw.model.job import Job, JobExecution
from aw.model.permission import CHOICE_PERMISSION_READ, CHOICE_PERMISSION_EXECUTE, \
    CHOICE_PERMISSION_WRITE, CHOICE_PERMISSION_DELETE
from aw.model.job_credential import JobGlobalCredentials
from aw.api_endpoints.base import API_PERMISSION, get_api_user, BaseResponse, GenericResponse, \
    LogDownloadResponse
from aw.api_endpoints.job_util import get_viewable_jobs_serialized, JobReadResponse, get_job_executions_serialized, \
    JobExecutionReadResponse, get_viewable_jobs, get_job_execution_serialized, get_log_file_content
from aw.utils.permission import has_job_permission, has_credentials_permission, has_manager_privileges
from aw.execute.queue import queue_add
from aw.execute.util import update_status, is_execution_status
from aw.utils.util import is_set
from aw.base import USERS


class JobWriteRequest(serializers.ModelSerializer):
    class Meta:
        model = Job
        fields = Job.api_fields_write


def _find_job(job_id: int) -> (Job, None):
    try:
        return Job.objects.get(id=job_id)

    except ObjectDoesNotExist:
        return None


def _find_job_and_execution(job_id: int, exec_id: int) -> tuple[Job, (JobExecution, None)]:
    job = _find_job(job_id)
    try:
        return job, JobExecution.objects.get(id=exec_id, job=job)

    except ObjectDoesNotExist:
        return job, None


def _job_execution_count(request) -> (int, None):
    max_count = None
    if 'execution_count' in request.GET:
        max_count = int(request.GET['execution_count'])
        max_count = min(max_count, 1000)

    return max_count


def _want_job_executions(request) -> tuple:
    max_count = None
    if 'executions' in request.GET and request.GET['executions'] == 'true':
        try:
            return True, _job_execution_count(request)

        except TypeError:
            pass

    return False, max_count


def _has_credentials_permission(user: USERS, data: dict) -> bool:
    if 'credentials_default' in data and is_set(data['credentials_default']):
        try:
            credentials = data['credentials_default']
            if not isinstance(credentials, JobGlobalCredentials):
                credentials = JobGlobalCredentials.objects.get(id=credentials)

            return has_credentials_permission(
                user=user,
                credentials=credentials,
                permission_needed=CHOICE_PERMISSION_READ,
            )

        except ObjectDoesNotExist:
            pass

    return True


class APIJob(APIView):
    http_method_names = ['post', 'get']
    serializer_class = JobReadResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobReadResponse, description='Return list of jobs'),
        },
        summary='Return list of all jobs the current user is privileged to view.',
        operation_id='job_list',
        parameters=[
            OpenApiParameter(
                name='executions', type=bool, default=False, description='Return list of job-executions',
                required=False,
            ),
            OpenApiParameter(
                name='execution_count', type=int, default=JOB_EXECUTION_LIMIT,
                description='Maximum count of job-executions to return',
                required=False,
            ),
        ],
    )
    def get(request):
        want_exec, exec_count = _want_job_executions(request)
        if want_exec:
            data = get_viewable_jobs_serialized(
                user=get_api_user(request),
                executions=True,
                execution_count=exec_count,
            )

        else:
            data = get_viewable_jobs_serialized(get_api_user(request))

        return Response(data=data, status=200)

    @extend_schema(
        request=JobWriteRequest,
        responses={
            200: OpenApiResponse(GenericResponse, description='Job created'),
            400: OpenApiResponse(GenericResponse, description='Invalid job data provided'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to create jobs'),
        },
        summary='Create a new job.',
        operation_id='job_create'
    )
    def post(self, request):
        user = get_api_user(request)
        if not has_manager_privileges(user=user, kind='job'):
            return Response(data={'msg': 'Not privileged to create jobs'}, status=403)

        serializer = JobWriteRequest(data=request.data)

        if not serializer.is_valid():
            return Response(
                data={'msg': f"Provided job data is not valid: '{serializer.errors}'"},
                status=400,
            )

        if not _has_credentials_permission(user=user, data=serializer.validated_data):
            return Response(
                data={'msg': "Not privileged to use provided credentials"},
                status=403,
            )

        try:
            serializer.save()

        except IntegrityError as err:
            return Response(
                data={'msg': f"Provided job data is not valid: '{err}'"},
                status=400,
            )

        return Response(data={'msg': 'Job created'}, status=200)


class APIJobItem(APIView):
    http_method_names = ['get', 'delete', 'put', 'post']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobReadResponse, description='Return job information'),
            400: OpenApiResponse(GenericResponse, description='Bad parameters provided'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to view the job'),
            404: OpenApiResponse(JobReadResponse, description='Job does not exist'),
        },
        summary='Return information about a job.',
        operation_id='job_view',
        parameters=[
            OpenApiParameter(
                name='executions', type=bool, default=False, description='Return list of job-executions',
                required=False,
            ),
            OpenApiParameter(
                name='execution_count', type=int, default=JOB_EXECUTION_LIMIT,
                description='Maximum count of job-executions to return',
                required=False,
            ),
        ],
    )
    def get(self, request, job_id: int):
        self.serializer_class = JobReadResponse
        user = get_api_user(request)
        job = _find_job(job_id)
        if job is None:
            return Response(data={'msg': f"Job with ID {job_id} does not exist"}, status=404)

        if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_READ):
            return Response(data={'msg': f"Job '{job.name}' is not viewable"}, status=403)

        data = JobReadResponse(instance=job).data

        want_exec, exec_count = _want_job_executions(request)
        if want_exec:
            data['executions'] = get_job_executions_serialized(job=job, execution_count=exec_count)

        return Response(data=data, status=200)

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(GenericResponse, description='Job deleted'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to delete the job'),
            404: OpenApiResponse(GenericResponse, description='Job does not exist'),
        },
        summary='Delete a job.',
        operation_id='job_delete'
    )
    def delete(self, request, job_id: int):
        user = get_api_user(request)
        try:
            job = _find_job(job_id)

            if job is not None:
                if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_DELETE):
                    return Response(data={'msg': f"Not privileged to delete the job '{job.name}'"}, status=403)

                job.delete()
                return Response(data={'msg': f"Job '{job.name}' deleted"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Job with ID {job_id} does not exist"}, status=404)

    @extend_schema(
        request=JobWriteRequest,
        responses={
            200: OpenApiResponse(GenericResponse, description='Job updated'),
            400: OpenApiResponse(GenericResponse, description='Invalid job data provided'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to modify the job'),
            404: OpenApiResponse(GenericResponse, description='Job does not exist'),
        },
        summary='Modify a job.',
        operation_id='job_edit'
    )
    def put(self, request, job_id: int):
        user = get_api_user(request)
        try:
            job = _find_job(job_id)

            if job is not None:
                if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_WRITE):
                    return Response(data={'msg': f"Not privileged to modify the job '{job.name}'"}, status=403)

                serializer = JobWriteRequest(data=request.data)
                if not serializer.is_valid():
                    return Response(
                        data={'msg': f"Provided job data is not valid: '{serializer.errors}'"},
                        status=400,
                    )

                if not _has_credentials_permission(user=user, data=serializer.validated_data):
                    return Response(
                        data={'msg': "Not privileged to use provided credentials"},
                        status=403,
                    )

                try:
                    Job.objects.filter(id=job.id).update(**serializer.data)

                except IntegrityError as err:
                    return Response(
                        data={'msg': f"Provided job data is not valid: '{err}'"},
                        status=400,
                    )

                return Response(data={'msg': f"Job '{job.name}' updated"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Job with ID {job_id} does not exist"}, status=404)

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobReadResponse, description='Job execution queued'),
            403: OpenApiResponse(JobReadResponse, description='Not privileged to execute the job'),
            404: OpenApiResponse(JobReadResponse, description='Job does not exist'),
        },
        summary='Execute a job.',
        operation_id='job_execute'
    )
    def post(self, request, job_id: int):
        user = get_api_user(request)
        try:
            job = _find_job(job_id)

            if job is not None:
                if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_EXECUTE):
                    return Response(data={'msg': f"Not privileged to execute the job '{job.name}'"}, status=403)

                queue_add(job=job, user=user)
                return Response(data={'msg': f"Job '{job.name}' execution queued"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Job with ID '{job_id}' does not exist"}, status=404)


class APIJobExecutionItem(APIView):
    http_method_names = ['delete']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobReadResponse, description='Job execution stopping'),
            400: OpenApiResponse(JobReadResponse, description='Job execution is not running'),
            403: OpenApiResponse(JobReadResponse, description='Not privileged to stop the job'),
            404: OpenApiResponse(JobReadResponse, description='Job or execution does not exist'),
        },
        summary='Stop a running job execution.',
        operation_id='job_exec_stop'
    )
    def delete(self, request, job_id: int, exec_id: int):
        user = get_api_user(request)
        try:
            job, execution = _find_job_and_execution(job_id, exec_id)

            if job is not None and execution is not None:
                if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_EXECUTE):
                    return Response(data={'msg': f"Not privileged to stop the job '{job.name}'"}, status=403)

                if not is_execution_status(execution, 'Running'):
                    return Response(data={'msg': f"Job execution '{job.name}' is not running"}, status=400)

                update_status(execution, 'Stopping')
                return Response(data={'msg': f"Job execution '{job.name}' stopping"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Job with ID '{job_id}' or execution does not exist"}, status=404)


class JobExecutionLogReadResponse(BaseResponse):
    lines = serializers.ListSerializer(child=serializers.CharField())


class APIJobExecutionLogs(APIView):
    http_method_names = ['get']
    serializer_class = JobExecutionLogReadResponse
    permission_classes = API_PERMISSION

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobExecutionLogReadResponse, description='Return job logs'),
            403: OpenApiResponse(JobExecutionLogReadResponse, description='Not privileged to view the job logs'),
            404: OpenApiResponse(JobExecutionLogReadResponse, description='Job, execution or log-file do not exist'),
        },
        summary='Get logs of a job execution.',
        operation_id='job_exec_logs'
    )
    def get(self, request, job_id: int, exec_id: int, line_start: int = 0):
        user = get_api_user(request)
        try:
            job, execution = _find_job_and_execution(job_id, exec_id)

            if job is not None and execution is not None:
                if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_READ):
                    return Response(data={'msg': f"Not privileged to view logs of the job '{job.name}'"}, status=403)

                if execution.log_stdout is None:
                    return Response(data={'msg': f"No logs found for job '{job.name}'"}, status=404)

                with open(execution.log_stdout, 'r', encoding='utf-8') as logfile:
                    lines = logfile.readlines()
                    return Response(data={'lines': lines[line_start:]}, status=200)

        except (ObjectDoesNotExist, FileNotFoundError):
            pass

        return Response(
            data={'msg': f"Job with ID '{job_id}', execution with ID '{exec_id}' or log-file does not exist"},
            status=404,
        )


class APIJobExecutionLogFile(APIView):
    http_method_names = ['get']
    serializer_class = LogDownloadResponse
    permission_classes = API_PERMISSION
    valid_logfile_type = ['stdout', 'stderr', 'stdout_repo', 'stderr_repo']

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(GenericResponse, description='Download job log-file'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to view the job logs'),
            404: OpenApiResponse(GenericResponse, description='Job, execution or log-file do not exist'),
        },
        summary='Download log-file of a job execution.',
        operation_id='job_exec_logfile',
        parameters=[
            OpenApiParameter(
                name='type', type=str, default='stdout',
                description=f"Type of log-file to download. One of {valid_logfile_type}",
                required=False,
            ),
        ],
    )
    def get(self, request, job_id: int, exec_id: int):
        user = get_api_user(request)
        try:
            job, execution = _find_job_and_execution(job_id, exec_id)

            if job is not None and execution is not None:
                if not has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_READ):
                    return Response(data={'msg': f"Not privileged to view logs of the job '{job.name}'"}, status=403)

                logfile = execution.log_stdout
                if 'type' in request.GET:
                    logfile_type = request.GET['type'] if request.GET['type'] in self.valid_logfile_type else 'stdout'
                    logfile = getattr(execution, f'log_{logfile_type}')

                if logfile is None:
                    return Response(data={'msg': f"No logs found for job '{job.name}'"}, status=404)

                return get_log_file_content(logfile)

        except (ObjectDoesNotExist, FileNotFoundError):
            pass

        return Response(
            data={'msg': f"Job with ID '{job_id}', execution with ID '{exec_id}' or log-file does not exist"},
            status=404,
        )


class APIJobExecution(APIView):
    http_method_names = ['get']
    serializer_class = JobExecutionReadResponse
    permission_classes = API_PERMISSION

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobExecutionReadResponse, description='Return job-execution information'),
        },
        summary='Return list of job-executions the current user is privileged to view.',
        operation_id='job_exec_list',
        parameters=[
            OpenApiParameter(
                name='execution_count', type=int, default=JOB_EXECUTION_LIMIT,
                description='Maximum count of job-executions to return',
                required=False,
            ),
        ],
    )
    def get(self, request):
        jobs = get_viewable_jobs(get_api_user(request))
        exec_count = _job_execution_count(request)
        if exec_count is None:
            exec_count = JOB_EXECUTION_LIMIT

        serialized = []
        for execution in JobExecution.objects.filter(job__in=jobs).order_by('-updated')[:exec_count]:
            serialized.append(get_job_execution_serialized(execution))

        return Response(data=serialized, status=200)



from django.db import models

from aw.model.base import BareModel, BaseModel, CHOICES_BOOL
from aw.model.job import Job
from aw.model.job_credential import JobGlobalCredentials
from aw.model.repository import Repository
from aw.base import USERS, GROUPS
from aw.utils.util import get_choice_value_by_key

CHOICE_PERMISSION_READ = 5
CHOICE_PERMISSION_EXECUTE = 10
CHOICE_PERMISSION_WRITE = 15
CHOICE_PERMISSION_FULL = 20
CHOICE_PERMISSION_DELETE = CHOICE_PERMISSION_FULL
CHOICES_PERMISSION = [
    (0, 'None'),
    (CHOICE_PERMISSION_READ, 'Read'),
    (CHOICE_PERMISSION_EXECUTE, 'Execute'),
    (CHOICE_PERMISSION_WRITE, 'Write'),
    (CHOICE_PERMISSION_FULL, 'Full'),
]


class JobPermission(BaseModel):
    form_fields = [
        'name', 'permission', 'users', 'groups', 'jobs', 'jobs_all', 'credentials', 'credentials_all',
        'repositories', 'repositories_all',
    ]
    api_fields_write = form_fields
    api_fields_read = ['permission_name', 'jobs_name', 'credentials_name', 'users_name', 'groups_name']
    api_fields_read.extend(form_fields)

    name = models.CharField(max_length=100, null=False, blank=False)
    permission_default = 0
    permission = models.PositiveSmallIntegerField(choices=CHOICES_PERMISSION, default=permission_default)
    users = models.ManyToManyField(
        USERS,
        through='JobPermissionMemberUser',
        through_fields=('permission', 'user'),
    )
    groups = models.ManyToManyField(
        GROUPS,
        through='JobPermissionMemberGroup',
        through_fields=('permission', 'group'),
    )
    jobs = models.ManyToManyField(
        Job,
        through='JobPermissionMapping',
        through_fields=('permission', 'job'),
    )
    jobs_all = models.BooleanField(choices=CHOICES_BOOL, default=False)
    credentials = models.ManyToManyField(
        JobGlobalCredentials,
        through='JobCredentialsPermissionMapping',
        through_fields=('permission', 'credentials'),
    )
    credentials_all = models.BooleanField(choices=CHOICES_BOOL, default=False)
    repositories = models.ManyToManyField(
        Repository,
        through='JobRepositoryPermissionMapping',
        through_fields=('permission', 'repository'),
    )
    repositories_all = models.BooleanField(choices=CHOICES_BOOL, default=False)

    @property
    def permission_name(self) -> str:
        return self.permission_name_from_id(self.permission)

    @staticmethod
    def permission_name_from_id(permission) -> str:
        return get_choice_value_by_key(choices=CHOICES_PERMISSION, find=permission)

    def __str__(self) -> str:
        return f"Permission '{self.name}' - {self.permission_name}"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['name'], name='jobperm_name_unique')
        ]


class JobPermissionMapping(BareModel):
    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    permission = models.ForeignKey(JobPermission, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"Permission '{self.permission.name}' linked to job '{self.job.name}'"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['job', 'permission'], name='jobpermmap_unique')
        ]


class JobCredentialsPermissionMapping(BareModel):
    credentials = models.ForeignKey(JobGlobalCredentials, on_delete=models.CASCADE)
    permission = models.ForeignKey(JobPermission, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"Permission '{self.permission.name}' linked to credentials '{self.credentials.name}'"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['credentials', 'permission'], name='jobcredpermmap_unique')
        ]


class JobRepositoryPermissionMapping(BareModel):
    repository = models.ForeignKey(Repository, on_delete=models.CASCADE)
    permission = models.ForeignKey(JobPermission, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"Permission '{self.permission.name}' linked to repository '{self.repository.name}'"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['repository', 'permission'], name='jobrepopermmap_unique')
        ]


class JobPermissionMemberUser(BareModel):
    user = models.ForeignKey(USERS, on_delete=models.CASCADE)
    permission = models.ForeignKey(JobPermission, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"Permission '{self.permission.name}' member user '{self.user.username}'"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['user', 'permission'], name='jobpermmemberuser_unique')
        ]


class JobPermissionMemberGroup(BareModel):
    group = models.ForeignKey(GROUPS, on_delete=models.CASCADE)
    permission = models.ForeignKey(JobPermission, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"Permission '{self.permission.name}' member group '{self.group.name}'"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['group', 'permission'], name='jobpermmembergrp_unique')
        ]



from django.db import models

CHOICES_BOOL = (
    (True, 'Yes'),
    (False, 'No')
)
DEFAULT_NONE = {'null': True, 'default': None, 'blank': True}
CHOICES_JOB_EXEC_STATUS = [
    (0, 'Waiting'),
    (1, 'Starting'),
    (2, 'Running'),
    (3, 'Failed'),
    (4, 'Finished'),
    (5, 'Stopping'),
    (6, 'Stopped'),
]


class BareModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True


class BaseModel(BareModel):
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True



from django.db import models

from aw.model.base import BaseModel, DEFAULT_NONE
from aw.utils.util import is_null, is_set
from aw.utils.crypto import decrypt, encrypt
from aw.base import USERS


class BaseJobCredentials(BaseModel):
    SECRET_ATTRS = ['become_pass', 'vault_pass', 'connect_pass', 'ssh_key']
    SECRET_ATTRS_ARGS = {
        'vault_pass': 'vault-password-file',
        'become_pass': 'become-password-file',
        'connect_pass': 'connection-password-file',
        'ssh_key': 'key-file',
    }
    PUBLIC_ATTRS_ARGS = {
        'connect_user': '--user',
        'become_user': '--become-user',
        'vault_file': '--vault-password-file',
        'vault_id': '--vault-id',
    }
    SECRET_HIDDEN = '⬤' * 15

    name = models.CharField(max_length=100, null=False, blank=False)
    connect_user = models.CharField(max_length=100, **DEFAULT_NONE)
    become_user = models.CharField(max_length=100, default='root', null=True, blank=True)
    # default become_user according to ansible-playbook docs
    vault_file = models.CharField(max_length=300, **DEFAULT_NONE)
    vault_id = models.CharField(max_length=50, **DEFAULT_NONE)

    _enc_vault_pass = models.CharField(max_length=500, **DEFAULT_NONE)
    _enc_become_pass = models.CharField(max_length=500, **DEFAULT_NONE)
    _enc_connect_pass = models.CharField(max_length=500, **DEFAULT_NONE)
    _enc_ssh_key = models.CharField(max_length=5000, **DEFAULT_NONE)

    @property
    def vault_pass(self) -> str:
        if is_null(self._enc_vault_pass):
            return ''

        return decrypt(self._enc_vault_pass)

    @vault_pass.setter
    def vault_pass(self, value: str):
        if is_null(value):
            self._enc_vault_pass = None
            return

        self._enc_vault_pass = encrypt(value)

    @property
    def vault_pass_is_set(self) -> bool:
        return is_set(self._enc_vault_pass)

    @property
    def become_pass(self) -> str:
        if is_null(self._enc_become_pass):
            return ''

        return decrypt(self._enc_become_pass)

    @become_pass.setter
    def become_pass(self, value: str):
        if is_null(value):
            self._enc_become_pass = None
            return

        self._enc_become_pass = encrypt(value)

    @property
    def become_pass_is_set(self) -> bool:
        return is_set(self._enc_become_pass)

    @property
    def connect_pass(self) -> str:
        if is_null(self._enc_connect_pass):
            return ''

        return decrypt(self._enc_connect_pass)

    @connect_pass.setter
    def connect_pass(self, value: str):
        if is_null(value):
            self._enc_connect_pass = None
            return

        self._enc_connect_pass = encrypt(value)

    @property
    def connect_pass_is_set(self) -> bool:
        return is_set(self._enc_connect_pass)

    @property
    def ssh_key(self) -> str:
        if is_null(self._enc_ssh_key):
            return ''

        return decrypt(self._enc_ssh_key)

    @ssh_key.setter
    def ssh_key(self, value: str):
        if is_null(value):
            self._enc_ssh_key = None
            return

        self._enc_ssh_key = encrypt(value)

    @property
    def ssh_key_is_set(self) -> bool:
        return is_set(self._enc_ssh_key)

    def _get_set_creds_str(self) -> str:
        creds_set = [attr for attr in self.SECRET_ATTRS if is_set(getattr(self, attr))]
        creds_set_str = ''
        if len(creds_set) > 0:
            creds_set_str = f" ({', '.join(creds_set)})"

        return creds_set_str

    class Meta:
        abstract = True


class JobGlobalCredentials(BaseJobCredentials):
    api_fields_read = [
        'id', 'name', 'become_user', 'connect_user', 'vault_file', 'vault_id'
    ]
    api_fields_write = api_fields_read.copy()
    api_fields_write.extend(BaseJobCredentials.SECRET_ATTRS)
    for secret_attr in BaseJobCredentials.SECRET_ATTRS:
        api_fields_read.append(f'{secret_attr}_is_set')
    form_fields = api_fields_write

    def __str__(self) -> str:
        return f"Global credentials '{self.name}'{self._get_set_creds_str()}"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['name'], name='jobcreds_name')
        ]


class JobUserCredentials(BaseJobCredentials):
    api_fields_read = JobGlobalCredentials.api_fields_read.copy()
    api_fields_read.append('user')
    api_fields_write = JobGlobalCredentials.api_fields_write.copy()
    api_fields_write.append('user')
    form_fields = JobGlobalCredentials.api_fields_write.copy()

    user = models.ForeignKey(USERS, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"Credentials '{self.name}' of user '{self.user.username}'{self._get_set_creds_str()}"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['user', 'name'], name='jobusercreds_user_name')
        ]



from pathlib import Path
from random import choice as random_choice
from string import digits, ascii_letters, punctuation
from datetime import datetime
from os import remove as remove_file
from re import sub as regex_replace

from aw.config.main import config
from aw.config.hardcoded import FILE_TIME_FORMAT
from aw.utils.util import write_file_0600, is_null, datetime_w_tz
from aw.utils.handlers import AnsibleConfigError
from aw.model.job import JobExecution, Job
from aw.model.repository import Repository


def config_error(msg: str):
    raise AnsibleConfigError(msg).with_traceback(None) from None


def overwrite_and_delete_file(file: (str, Path)):
    if not isinstance(file, Path):
        file = Path(file)

    if not file.is_file():
        return

    for _ in range(3):
        write_file_0600(
            file=file,
            content=''.join(random_choice(ascii_letters + digits + punctuation) for _ in range(50)),
        )

    remove_file(file)


def decode_job_env_vars(env_vars_csv: str, src: str) -> dict:
    try:
        env_vars = {}
        for kv in env_vars_csv.split(','):
            k, v = kv.split('=')
            env_vars[k] = v

        return env_vars

    except ValueError:
        config_error(
            f"Environmental variables of {src} are not in a valid format "
            f"(comma-separated key-value pairs). Example: 'key1=val1,key2=val2'"
        )
        return {}


def update_status(obj: (JobExecution, Repository), status: str):
    obj.status = obj.status_id_from_name(status)
    obj.save()


def is_execution_status(execution: JobExecution, status: str) -> bool:
    is_status = JobExecution.objects.get(id=execution.id).status
    check_status = execution.status_id_from_name(status)
    return is_status == check_status


def get_path_run() -> Path:
    # build unique temporary execution directory
    path_run = config['path_run']
    if not path_run.endswith('/'):
        path_run += '/'

    path_run += datetime.now().strftime(FILE_TIME_FORMAT)
    path_run += ''.join(random_choice(digits) for _ in range(5))
    return Path(path_run)


def create_dirs(path: (str, Path), desc: str):
    try:
        if not isinstance(path, Path):
            path = Path(path)

        path.mkdir(mode=0o750, parents=True, exist_ok=True)

    except (OSError, FileNotFoundError):
        raise OSError(f"Unable to created {desc} directory: '{path}'").with_traceback(None) from None


def job_logs(job: Job, execution: JobExecution) -> dict:
    safe_job_name = regex_replace(pattern='[^0-9a-zA-Z-_]+', repl='', string=job.name)
    if is_null(execution.user):
        safe_user_name = 'scheduled'
    else:
        safe_user_name = execution.user.username.replace('.', '_')
        safe_user_name = regex_replace(pattern='[^0-9a-zA-Z-_]+', repl='', string=safe_user_name)

    timestamp = datetime_w_tz().strftime(FILE_TIME_FORMAT)
    log_file = f"{config['path_log']}/{safe_job_name}_{timestamp}_{safe_user_name}"

    log_files = {
        'stdout': f'{log_file}_stdout.log',
        'stderr': f'{log_file}_stderr.log',
        'stdout_repo': f'{log_file}_stdout_repo.log',
        'stderr_repo': f'{log_file}_stderr_repo.log',
    }

    execution.log_stdout = log_files['stdout']
    execution.log_stderr = log_files['stderr']
    execution.log_stdout_repo = log_files['stdout_repo']
    execution.log_stderr_repo = log_files['stderr_repo']

    return log_files



from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import serializers
from drf_spectacular.utils import extend_schema, OpenApiResponse
from django.db.utils import IntegrityError

from aw.config.main import config
from aw.model.system import SystemConfig, get_config_from_db
from aw.api_endpoints.base import API_PERMISSION, get_api_user, GenericResponse, BaseResponse
from aw.utils.util_no_config import is_set
from aw.utils.debug import log
from aw.utils.permission import has_manager_privileges


class SystemConfigReadResponse(BaseResponse):
    # todo: fix static fields.. duplicate logic in model

    # SystemConfig.form_fields
    path_run = serializers.CharField()
    path_play = serializers.CharField()
    path_log = serializers.CharField()
    timezone = serializers.CharField()
    run_timeout = serializers.IntegerField()
    session_timeout = serializers.IntegerField()
    path_ansible_config = serializers.CharField()
    path_ssh_known_hosts = serializers.CharField()

    # SystemConfig.api_fields_read_only
    db = serializers.CharField()
    db_migrate = serializers.BooleanField()
    serve_static = serializers.BooleanField()
    deployment = serializers.CharField()
    version = serializers.CharField()
    logo_url = serializers.CharField()
    ara_server = serializers.CharField()
    global_environment_vars = serializers.CharField()


class SystemConfigWriteRequest(serializers.ModelSerializer):
    class Meta:
        model = SystemConfig
        fields = SystemConfig.api_fields_write


class APISystemConfig(APIView):
    http_method_names = ['put', 'get']
    serializer_class = SystemConfigReadResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={200: SystemConfigReadResponse},
        summary='Return currently active config.',
        operation_id='system_config_view',
    )
    def get(request):
        del request
        merged_config = {'read_only_settings': SystemConfig.api_fields_read_only}

        for field in SystemConfig.form_fields + merged_config['read_only_settings']:
            merged_config[field] = config[field]

        merged_config['read_only_settings'] += SystemConfig.get_set_env_vars()

        return Response(merged_config)

    @extend_schema(
        request=SystemConfigWriteRequest,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='System config updated'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid system config provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to update the system config'),
        },
        summary='Modify system config.',
        operation_id='system_config_edit',
    )
    def put(self, request):
        privileged = has_manager_privileges(user=get_api_user(request), kind='system')
        if not privileged:
            return Response(
                data={'msg': 'Not privileged to manage system config'},
                status=403,
            )

        serializer = SystemConfigWriteRequest(data=request.data)

        if not serializer.is_valid():
            return Response(
                data={'msg': f"Provided system config is not valid: '{serializer.errors}'"},
                status=400,
            )

        config_db = get_config_from_db()
        try:
            changed = False
            for setting, value in serializer.validated_data.items():
                if is_set(value) and str(config[setting]) != str(value):
                    setattr(config_db, setting, value)
                    changed = True

            if changed:
                log(msg='System config change - updating', level=5)
                config_db.save()

            return Response(data={'msg': "System config updated"}, status=200)

        except IntegrityError as err:
            return Response(data={'msg': f"Provided system config is not valid: '{err}'"}, status=400)



from crontab import CronTab
from django.db import models
from django.core.validators import ValidationError
from django.utils import timezone

from aw.config.main import config
from aw.model.base import BareModel, BaseModel, CHOICES_BOOL, DEFAULT_NONE, CHOICES_JOB_EXEC_STATUS
from aw.config.hardcoded import SHORT_TIME_FORMAT
from aw.model.job_credential import JobGlobalCredentials, JobUserCredentials
from aw.base import USERS
from aw.model.repository import Repository
from aw.utils.util import get_choice_key_by_value, get_choice_value_by_key, datetime_from_db_str, is_null


class JobError(BareModel):
    short = models.CharField(max_length=100)
    med = models.TextField(max_length=1024, null=True)

    def __str__(self) -> str:
        return f"Job error {self.created}: '{self.short}'"


CHOICES_JOB_VERBOSITY = (
    (0, 'Default'),
    (1, 'v'),
    (2, 'vv'),
    (3, 'vvv'),
    (4, 'vvvv'),
    (5, 'vvvv'),
    (6, 'vvvvvv'),
)


class BaseJob(BaseModel):
    BAD_ANSIBLE_FLAGS = [
        'step', 'ask-vault-password', 'ask-vault-pass', 'k', 'ask-pass',
    ]

    limit = models.CharField(max_length=500, **DEFAULT_NONE)
    verbosity = models.PositiveSmallIntegerField(choices=CHOICES_JOB_VERBOSITY, default=0)
    comment = models.CharField(max_length=150, **DEFAULT_NONE)
    mode_diff = models.BooleanField(choices=CHOICES_BOOL, default=False)
    mode_check = models.BooleanField(choices=CHOICES_BOOL, default=False)

    # NOTE: one or multiple comma-separated vars
    environment_vars = models.CharField(max_length=1000, **DEFAULT_NONE)

    tags = models.CharField(max_length=150, **DEFAULT_NONE)
    tags_skip = models.CharField(max_length=150, **DEFAULT_NONE)
    cmd_args = models.CharField(max_length=150, **DEFAULT_NONE)

    class Meta:
        abstract = True

    def clean(self):
        super().clean()

        for flag in self.BAD_ANSIBLE_FLAGS:
            for search in [f'-{flag} ', f'-{flag}=', f'-{flag}']:
                if self.cmd_args.find(search) != -1:
                    raise ValidationError(
                        f"Found one or more bad flags in commandline arguments: {self.BAD_ANSIBLE_FLAGS} (prompts)"
                    )


def validate_cronjob(value):
    try:
        _ = CronTab(value)
        return value

    except ValueError:
        raise ValidationError('The provided schedule is not in a valid cron format')


class Job(BaseJob):
    CHANGE_FIELDS = [
        'name', 'playbook_file', 'inventory_file', 'repository', 'schedule', 'enabled', 'limit', 'verbosity',
        'mode_diff', 'mode_check', 'tags', 'tags_skip', 'verbosity', 'comment', 'environment_vars', 'cmd_args',
        'credentials_default', 'credentials_needed',
    ]
    form_fields_primary = ['name', 'playbook_file', 'inventory_file', 'repository']
    form_fields = CHANGE_FIELDS
    api_fields_read = ['id']
    api_fields_read.extend(CHANGE_FIELDS)
    api_fields_write = api_fields_read.copy()
    api_fields_read.append('next_run')

    name = models.CharField(max_length=150, null=False, blank=False)
    playbook_file = models.CharField(max_length=100)
    # NOTE: one or multiple comma-separated inventories
    inventory_file = models.CharField(max_length=300, **DEFAULT_NONE)
    schedule_max_len = 50
    schedule = models.CharField(max_length=schedule_max_len, validators=[validate_cronjob], **DEFAULT_NONE)
    enabled = models.BooleanField(choices=CHOICES_BOOL, default=True)

    credentials_needed = models.BooleanField(choices=CHOICES_BOOL, default=False)
    credentials_default = models.ForeignKey(
        JobGlobalCredentials, on_delete=models.SET_NULL, related_name='job_fk_creddflt', null=True, blank=True,
    )
    repository = models.ForeignKey(Repository, on_delete=models.SET_NULL, related_name='job_fk_repo', **DEFAULT_NONE)

    def __str__(self) -> str:
        limit = '' if self.limit is None else f' [{self.limit}]'
        return f"Job '{self.name}' ({self.playbook_file} => {self.inventory_file}{limit})"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['name'], name='job_name_unique')
        ]


class JobExecutionResult(BareModel):
    # ansible_runner.runner.Runner
    time_start = models.DateTimeField(default=timezone.now)
    time_fin = models.DateTimeField(**DEFAULT_NONE)

    failed = models.BooleanField(choices=CHOICES_BOOL, default=False)
    error = models.ForeignKey(JobError, on_delete=models.SET_NULL, related_name='jobresult_fk_error', null=True)

    def __str__(self) -> str:
        result = 'succeeded'

        if self.failed:
            result = 'failed'

        return f"Job execution {self.time_start}: {result}"

    @property
    def time_fin_str(self) -> str:
        return datetime_from_db_str(dt=self.time_fin, fmt=SHORT_TIME_FORMAT) + f" {config['timezone']}"


class JobExecutionResultHost(BareModel):
    STATS = [
        'unreachable', 'tasks_skipped', 'tasks_ok', 'tasks_failed', 'tasks_rescued',
        'tasks_ignored', 'tasks_changed',
    ]
    # ansible_runner.runner.Runner.stats
    hostname = models.CharField(max_length=300, null=False)
    unreachable = models.BooleanField(choices=CHOICES_BOOL, default=False)

    tasks_skipped = models.PositiveSmallIntegerField(default=0)
    tasks_ok = models.PositiveSmallIntegerField(default=0)
    tasks_failed = models.PositiveSmallIntegerField(default=0)
    tasks_rescued = models.PositiveSmallIntegerField(default=0)
    tasks_ignored = models.PositiveSmallIntegerField(default=0)
    tasks_changed = models.PositiveSmallIntegerField(default=0)

    error = models.ForeignKey(JobError, on_delete=models.SET_NULL, related_name='jobresulthost_fk_error', null=True)
    result = models.ForeignKey(
        JobExecutionResult, on_delete=models.CASCADE, related_name='jobresulthost_fk_result', null=True
    )

    def __str__(self) -> str:
        result = 'succeeded'

        if int(self.tasks_failed) > 0:
            result = 'failed'

        return f"Job execution {self.created} of host '{self.hostname}': {result}"


class JobExecution(BaseJob):
    api_fields_read = [
        'id', 'job', 'job_name', 'user', 'user_name', 'result', 'status', 'status_name', 'time_start', 'time_fin',
        'failed', 'error_s', 'error_m', 'log_stdout', 'log_stdout_url', 'log_stderr', 'log_stderr_url', 'job_comment',
        'credential_global', 'credential_user', 'command', 'log_stdout_repo', 'log_stderr_repo',
        'log_stdout_repo_url', 'log_stderr_repo_url',
    ]
    log_file_fields = ['log_stdout', 'log_stderr', 'log_stdout_repo', 'log_stderr_repo']

    # NOTE: scheduled execution will have no user
    user = models.ForeignKey(
        USERS, on_delete=models.SET_NULL, null=True,
        related_name='jobexec_fk_user', editable=False,
    )
    job = models.ForeignKey(Job, on_delete=models.CASCADE, related_name='jobexec_fk_job')
    result = models.ForeignKey(
        JobExecutionResult, on_delete=models.SET_NULL, related_name='jobexec_fk_result',
        **DEFAULT_NONE,  # execution is created before result is available
    )
    status = models.PositiveSmallIntegerField(default=0, choices=CHOICES_JOB_EXEC_STATUS)
    user_credentials = models.ForeignKey(
        Job, on_delete=models.SET_NULL, related_name='jobexec_fk_usercreds', **DEFAULT_NONE,
    )
    log_stdout = models.CharField(max_length=300, **DEFAULT_NONE)
    log_stderr = models.CharField(max_length=300, **DEFAULT_NONE)
    log_stdout_repo = models.CharField(max_length=300, **DEFAULT_NONE)
    log_stderr_repo = models.CharField(max_length=300, **DEFAULT_NONE)
    command = models.CharField(max_length=1000, **DEFAULT_NONE)

    credential_global = models.ForeignKey(
        JobGlobalCredentials, on_delete=models.SET_NULL, related_name='jobexec_fk_credglob', null=True,
    )
    credential_user = models.ForeignKey(
        JobUserCredentials, on_delete=models.SET_NULL, related_name='jobexec_fk_credusr', null=True,
    )

    def __str__(self) -> str:
        executor = 'scheduled'
        if self.user is not None:
            executor = self.user.username

        return f"Job '{self.job.name}' execution @ {self.time_created_str} by '{executor}': {self.status_name}"

    @property
    def status_name(self) -> str:
        return self.status_name_from_id(self.status)

    @staticmethod
    def status_name_from_id(rtype) -> str:
        return get_choice_value_by_key(choices=CHOICES_JOB_EXEC_STATUS, find=rtype)

    @staticmethod
    def status_id_from_name(name: str) -> int:
        return get_choice_key_by_value(choices=CHOICES_JOB_EXEC_STATUS, find=name)

    @property
    def time_created_str(self) -> str:
        if is_null(self.created):
            return ''

        return datetime_from_db_str(dt=self.created, fmt=SHORT_TIME_FORMAT) + f" {config['timezone']}"

    @property
    def log_stdout_url(self) -> str:
        return f"/api/job/{self.job.id}/{self.id}/log?type=stdout"

    @property
    def log_stderr_url(self) -> str:
        return f"/api/job/{self.job.id}/{self.id}/log?type=stderr"

    @property
    def log_stdout_repo_url(self) -> str:
        return f"/api/job/{self.job.id}/{self.id}/log?type=stdout_repo"

    @property
    def log_stderr_repo_url(self) -> str:
        return f"/api/job/{self.job.id}/{self.id}/log?type=stderr_repo"


class JobQueue(BareModel):
    job = models.ForeignKey(Job, on_delete=models.CASCADE, related_name='jobqueue_fk_job')
    user = models.ForeignKey(
        USERS, on_delete=models.SET_NULL, null=True,
        related_name='jobqueue_fk_user',
    )



from django.db import models
from django.core.exceptions import ObjectDoesNotExist

from aw.model.base import BaseModel, CHOICES_BOOL, DEFAULT_NONE
from aw.config.defaults import CONFIG_DEFAULTS
from aw.config.environment import check_aw_env_var_is_set
from aw.config.main import VERSION


# NOTE: add default-values to config.defaults.CONFIG_DEFAULTS
class SystemConfig(BaseModel):
    form_fields = [
        'path_run', 'path_play', 'path_log', 'timezone', 'run_timeout', 'session_timeout', 'path_ansible_config',
        'path_ssh_known_hosts', 'debug', 'logo_url', 'ara_server', 'global_environment_vars',
    ]
    # NOTE: 'AW_DB' is needed to get this config from DB and 'AW_SECRET' cannot be saved because of security breach
    api_fields_write = form_fields
    api_fields_read_only = ['db', 'db_migrate', 'serve_static', 'deployment', 'version']

    path_run = models.CharField(max_length=500, default='/tmp/ansible-webui')
    path_play = models.CharField(max_length=500, default=None)
    path_log = models.CharField(max_length=500, default=None)
    timezone = models.CharField(max_length=300, default='UTC')  # UTC to keep model migrations static
    run_timeout = models.PositiveIntegerField(default=CONFIG_DEFAULTS['run_timeout'])
    session_timeout = models.PositiveIntegerField(default=CONFIG_DEFAULTS['session_timeout'])
    path_ansible_config = models.CharField(max_length=500, **DEFAULT_NONE)
    path_ssh_known_hosts = models.CharField(max_length=500, **DEFAULT_NONE)
    debug = models.BooleanField(default=False, choices=CHOICES_BOOL)
    logo_url = models.CharField(max_length=500, **DEFAULT_NONE)
    ara_server = models.CharField(max_length=300, **DEFAULT_NONE)
    global_environment_vars = models.CharField(max_length=1000, **DEFAULT_NONE)

    @classmethod
    def get_set_env_vars(cls) -> list:
        # grey-out settings in web-ui
        return [field for field in cls.form_fields if check_aw_env_var_is_set(field)]

    def __str__(self) -> str:
        return 'Ansible-WebUI System Config'


def get_config_from_db() -> SystemConfig:
    try:
        config_db = SystemConfig.objects.all().first()
        if config_db is None:
            raise ObjectDoesNotExist()

    except ObjectDoesNotExist:
        # create config-object and set dynamic defaults
        config_db = SystemConfig(
            path_play=CONFIG_DEFAULTS['path_play'],
            path_log=CONFIG_DEFAULTS['path_log'],
            timezone=CONFIG_DEFAULTS['timezone'],
        )
        config_db.save()

    return config_db


class SchemaMetadata(BaseModel):
    schema_version = models.CharField(max_length=50)
    schema_version_prev = models.CharField(max_length=50, **DEFAULT_NONE)


def get_schema_metadata() -> SchemaMetadata:
    try:
        metadata = SchemaMetadata.objects.all().first()
        if metadata is None:
            raise ObjectDoesNotExist()

    except ObjectDoesNotExist:
        metadata = SchemaMetadata(
            schema_version=VERSION,
        )
        metadata.save()

    return metadata



AW_VALIDATIONS = {
    'file_system_browse': ['inventory_file', 'playbook_file'],
    'file_system_exists': [
        'vault_file', 'static_path', 'path_run', 'path_play', 'path_log', 'path_ansible_config',
        'path_ssh_known_hosts',
    ],
}



from pathlib import Path
from shutil import rmtree
from re import sub as regex_replace

from django.utils import timezone

from aw.config.main import config
from aw.config.hardcoded import REPO_CLONE_TIMEOUT
from aw.model.job import Job, JobExecution
from aw.utils.util import is_null, is_set, write_file_0640
from aw.utils.subps import process
from aw.execute.play_credentials import write_pwd_file, get_pwd_file
from aw.execute.util import overwrite_and_delete_file, update_status, get_path_run, job_logs
from aw.model.job_credential import BaseJobCredentials
from aw.utils.handlers import AnsibleRepositoryError
from aw.model.repository import Repository
from aw.base import USERS


class ExecuteRepository:
    def __init__(self, repository: Repository, execution: JobExecution = None, path_run: Path = None):
        self.repository = repository
        self.path_run = path_run
        self.execution = execution
        self.path_repo = None

    def create_repository(self, env: dict):
        if is_set(self.repository.git_override_initialize):
            self._run_repo_config_cmds(
                cmds=self.repository.git_override_initialize,
                env=env,
            )
            return

        git_clone = ['git', 'clone', '--branch', self.repository.git_branch]

        if is_set(self.repository.git_limit_depth):
            git_clone.extend(['--depth', str(self.repository.git_limit_depth)])

        if self.path_repo is None:
            self.path_repo = self.get_path_repo()

        git_clone.extend([self._git_origin_with_credentials(), str(self.path_repo)])

        git_cmds = [' '.join(git_clone)]

        if self.repository.git_lfs:
            git_cmds.append('git lfs fetch')
            git_cmds.append('git lfs checkout')

        for cmd in git_cmds:
            self._repo_process(cmd=cmd, env=env)

    def update_repository(self, env: dict):
        if is_set(self.repository.git_override_update):
            self._run_repo_config_cmds(
                cmds=self.repository.git_override_update,
                env=env,
            )
            return

        git_pull = ['git', 'pull']

        if is_set(self.repository.git_limit_depth):
            git_pull.extend(['--depth', str(self.repository.git_limit_depth)])

        git_cmds = [
            'git reset --hard',
            ' '.join(git_pull),
        ]

        if self.repository.git_lfs:
            git_cmds.append('git lfs fetch')
            git_cmds.append('git lfs checkout')

        for cmd in git_cmds:
            self._repo_process(cmd=cmd, env=env)

    def create_or_update_repository(self):
        if is_null(self.repository) or self.repository.rtype_name == 'Static':
            return

        if self.execution is not None:
            self.repository.log_stderr = self.execution.log_stderr_repo
            self.repository.log_stdout = self.execution.log_stdout_repo
            self.repository.save()

        try:
            update_status(self.repository, status='Running')
            path_repo = self.get_path_repo()

            env = self._git_env()
            if len(env) > 0:
                self._log_file_write(f"USING ENVIRONMENT: {env}")

            self._run_repo_config_cmds(cmds=self.repository.git_hook_pre, env=env)

            if self.repository.git_isolate or not (Path(path_repo) / '.git/HEAD').is_file():
                self.create_repository(env=env)

            else:
                self.update_repository(env=env)

            self.repository.time_update = timezone.now()

            self._run_repo_config_cmds(cmds=self.repository.git_hook_post, env=env)

            update_status(self.repository, status='Finished')

        # pylint: disable=W0718
        except Exception as err:
            self._error(msg=f"Got unexpected error: '{err}'")

    def _error(self, msg: str):
        write_file_0640(file=self.repository.log_stderr, content=msg)
        update_status(self.repository, status='Failed')
        raise AnsibleRepositoryError(msg).with_traceback(None) from None

    def get_path_run_repo(self) -> Path:
        return self.path_run / '.repository'

    def _git_env(self) -> dict:
        env = {'GIT_SSH_COMMAND': 'ssh'}
        if self.repository.git_origin.find(' -p') != -1:
            try:
                self.repository.git_origin, port = self.repository.git_origin.split(' -p')
                env['GIT_SSH_COMMAND'] += f" -p {port}"

            except IndexError:
                pass

        if is_set(self.repository.git_credentials) and is_set(self.repository.git_credentials.ssh_key):
            path_run_repo = self.get_path_run_repo()
            path_run_repo.mkdir(mode=0o700, parents=True, exist_ok=True)
            write_pwd_file(credentials=self.repository.git_credentials, attr='ssh_key', path_run=path_run_repo)
            env['GIT_SSH_COMMAND'] += f" -i {get_pwd_file(path_run=path_run_repo, attr='ssh_key')}"

        if is_set(config['path_ssh_known_hosts']):
            if Path(config['path_ssh_known_hosts']).is_file():
                env['GIT_SSH_COMMAND'] += f" -o UserKnownHostsFile={config['path_ssh_known_hosts']}"

            else:
                self._log_file_write('Ignoring known_hosts file because it does not exist')

        if env['GIT_SSH_COMMAND'] == 'ssh':
            env.pop('GIT_SSH_COMMAND')

        return env

    def get_project_dir(self) -> str:
        if is_null(self.repository):
            return config['path_play']

        return str(self.get_path_playbook_base())

    def cleanup_repository(self):
        if is_null(self.repository) or self.repository.rtype_name == 'Static':
            return

        path_run_repo = self.get_path_run_repo()
        for attr in BaseJobCredentials.SECRET_ATTRS:
            overwrite_and_delete_file(get_pwd_file(path_run=path_run_repo, attr=attr))

        if self.repository.git_isolate:
            rmtree(self.get_path_repo(), ignore_errors=True)

    def get_path_repo(self) -> Path:
        path_repo = get_path_repo_wo_isolate(self.repository)

        if self.repository.git_isolate:
            path_repo = path_repo / self.execution.id
            path_repo.mkdir(mode=0o750, parents=True, exist_ok=True)

        return path_repo

    def get_path_playbook_base(self) -> Path:
        path_repo = self.get_path_repo()
        if self.repository.rtype_name == 'Git' and is_set(self.repository.git_playbook_base):
            path_repo = path_repo / self.repository.git_playbook_base

        return path_repo

    def _repo_process(self, cmd: str, env: dict):
        if self.path_repo is None:
            self.path_repo = self.get_path_repo()

        result = process(cmd=cmd, cwd=self.path_repo, env=env, shell=True, timeout_sec=REPO_CLONE_TIMEOUT)
        self._log_file_write(f"COMMAND: {cmd}\n{result['stdout']}")
        if result['rc'] != 0:
            self._error(
                f"Repository command failed: '{cmd}'\n"
                f"Got error: '{result['stderr']}'\n"
                f"Got output: '{result['stdout']}'"
            )

    def _run_repo_config_cmds(self, cmds: str, env: dict):
        if is_set(cmds):
            for cmd in cmds.split(','):
                self._repo_process(cmd=cmd, env=env)

    def _git_origin_with_credentials(self) -> str:
        origin = self.repository.git_origin

        if is_set(self.repository.git_credentials):
            credentials = self.repository.git_credentials

            if origin.find('://') != -1 and origin.find('ssh://') == -1:
                # not ssh
                if origin.find('@') == -1 and \
                    is_set(credentials.connect_user) and \
                        is_set(credentials.connect_pass):
                    proto, origin_host = origin.split('://')
                    origin = f'{proto}://{credentials.connect_user}:{credentials.connect_pass}@{origin_host}'

            else:
                # ssh
                if origin.find('@') == -1 and is_set(credentials.connect_user):
                    origin = f'{credentials.connect_user}@{origin}'

        return origin

    def _log_file_write(self, content: str):
        write_file_0640(
            file=self.repository.log_stdout,
            content=f"{content}\n"
        )


def get_path_repo_wo_isolate(repository: Repository) -> Path:
    if repository.rtype_name == 'Static':
        return repository.static_path

    safe_repo_name = regex_replace(pattern='[^0-9a-zA-Z-_]+', repl='', string=repository.name)
    path_repo = Path(config['path_run']) / 'repositories' / safe_repo_name
    path_repo.mkdir(mode=0o750, parents=True, exist_ok=True)
    return path_repo


def api_update_repository(repository: Repository, user: USERS):
    if is_null(repository) or repository.rtype_name == 'Static' or repository.git_isolate:
        return

    job = Job(name='RepoUpdate')
    execution = JobExecution(user=user, job=job)
    job_logs(job=job, execution=execution)

    repository.log_stderr = execution.log_stderr_repo
    repository.log_stdout = execution.log_stdout_repo
    repository.save()

    ExecuteRepository(
        repository=repository, path_run=get_path_run(),
    ).create_or_update_repository()



"""
used to serve static files if no proxy is in use
source: django.contrib.staticfiles.views(serve)

can be switched off by setting the environmental variable 'AW_STATIC'
"""
from posixpath import normpath
from os import path as os_path
from re import escape as regex_escape

from django.contrib.staticfiles import finders
from django.http import Http404
from django.views import static
from django.conf import settings
from django.urls import re_path


def serve(request, path, **kwargs):
    normalized_path = normpath(path).lstrip('/')
    absolute_path = finders.find(normalized_path)

    if not absolute_path:
        if path.endswith("/") or path == "":
            raise Http404('Directory indexes are not allowed here.')
        raise Http404(f"'{path}' could not be found")

    document_root, path = os_path.split(absolute_path)
    return static.serve(request, path, document_root=document_root, **kwargs)


# pylint: disable=C0209
urlpatterns_static = [re_path(
    r"^%s(?P<path>.*)$" % regex_escape(settings.STATIC_URL.lstrip('/')), serve,
)]



from base64 import b64encode, b64decode
from hashlib import sha256

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

from aw.config.main import config
from aw.utils.util import is_null
from aw.utils.debug import log_warn, log

__KEY = sha256(config['secret'].encode('utf-8')).digest()


def encrypt(plaintext: str) -> str:
    if is_null(plaintext):
        return ''

    try:
        return _encrypt(plaintext.encode('utf-8')).decode('utf-8')

    except ValueError as err:
        log_warn("Unable to encrypt data!")
        log(msg=f"Got error encrypting plaintext: '{err}'", level=6)
        return ''


def _encrypt(plaintext: bytes) -> bytes:
    iv = get_random_bytes(AES.block_size)
    cipher = AES.new(__KEY, AES.MODE_CBC, iv)
    ciphertext = iv + cipher.encrypt(
        plaintext=pad(
            data_to_pad=plaintext,
            block_size=AES.block_size,
            style='pkcs7',
        ),
    )
    return b64encode(ciphertext)


def decrypt(ciphertext: str) -> str:
    if is_null(ciphertext):
        return ''

    try:
        return _decrypt(ciphertext.encode('utf-8')).decode('utf-8')

    except ValueError as err:
        log_warn("Unable to decrypt secret! Maybe the key 'AW_SECRET' changed?")
        log(msg=f"Got error decrypting ciphertext: '{err}'", level=6)
        return ''


def _decrypt(ciphertext: bytes) -> bytes:
    ciphertext = b64decode(ciphertext)
    cipher = AES.new(__KEY, AES.MODE_CBC, ciphertext[:AES.block_size])
    return unpad(
        padded_data=cipher.decrypt(ciphertext[AES.block_size:]),
        block_size=AES.block_size,
        style='pkcs7',
    )



from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import IntegrityError
from rest_framework.generics import GenericAPIView
from rest_framework import serializers
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema, OpenApiResponse

from aw.model.job import Job
from aw.model.job_credential import JobGlobalCredentials
from aw.model.permission import JobPermission, JobPermissionMapping, JobPermissionMemberUser, \
    JobPermissionMemberGroup, JobCredentialsPermissionMapping, JobRepositoryPermissionMapping
from aw.api_endpoints.base import API_PERMISSION, GenericResponse, get_api_user
from aw.utils.permission import has_manager_privileges
from aw.utils.util import is_set
from aw.base import USERS, GROUPS
from aw.model.repository import Repository


class PermissionReadResponse(serializers.ModelSerializer):
    class Meta:
        model = JobPermission
        fields = JobPermission.api_fields_read

    permission_name = serializers.CharField(required=False)
    jobs = serializers.ListSerializer(child=serializers.IntegerField(), required=False)
    credentials = serializers.ListSerializer(child=serializers.IntegerField(), required=False)
    users_name = serializers.ListSerializer(child=serializers.CharField(), required=False)
    groups_name = serializers.ListSerializer(child=serializers.CharField(), required=False)
    jobs_name = serializers.ListSerializer(child=serializers.CharField(), required=False)
    credentials_name = serializers.ListSerializer(child=serializers.CharField(), required=False)


class JobSerializer(serializers.ModelSerializer):
    class Meta:
        model = Job


class PermissionWriteRequest(serializers.ModelSerializer):
    class Meta:
        model = JobPermission
        fields = JobPermission.api_fields_write

    jobs = serializers.MultipleChoiceField(allow_blank=True, choices=[])
    credentials = serializers.MultipleChoiceField(allow_blank=True, choices=[])
    repositories = serializers.MultipleChoiceField(allow_blank=True, choices=[])
    users = serializers.MultipleChoiceField(allow_blank=True, choices=[])
    groups = serializers.MultipleChoiceField(allow_blank=True, choices=[])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['jobs'] = serializers.MultipleChoiceField(choices=[job.id for job in Job.objects.all()])
        self.fields['credentials'] = serializers.MultipleChoiceField(
            choices=[creds.id for creds in JobGlobalCredentials.objects.all()]
        )
        self.fields['repositories'] = serializers.MultipleChoiceField(
            choices=[repo.id for repo in Repository.objects.all()]
        )
        self.fields['users'] = serializers.MultipleChoiceField(choices=[user.id for user in USERS.objects.all()])
        self.fields['groups'] = serializers.MultipleChoiceField(choices=[group.id for group in GROUPS.objects.all()])

    @staticmethod
    def create_or_update(validated_data: dict, perm: JobPermission = None):
        # pylint: disable=R0912,R0915
        if 'permission' in validated_data:
            permission = validated_data['permission']
        else:
            permission = JobPermission.permission_default

        if perm is None:
            perm = JobPermission(
                name=validated_data['name'],
                permission=permission,
            )

        else:
            perm.name = validated_data['name']
            perm.permission = permission

        if 'jobs_all' in validated_data:
            perm.jobs_all = validated_data['jobs_all']

        if 'credentials_all' in validated_data:
            perm.credentials_all = validated_data['credentials_all']

        if 'repositories_all' in validated_data:
            perm.repositories_all = validated_data['repositories_all']

        perm.save()

        if 'jobs' in validated_data and is_set(validated_data['jobs']):
            jobs = []
            for job_id in validated_data['jobs']:
                try:
                    jobs.append(Job.objects.get(id=job_id))

                except ObjectDoesNotExist:
                    continue

            perm.jobs.set(jobs)

        if 'credentials' in validated_data and is_set(validated_data['credentials']):
            credentials = []
            for creds_id in validated_data['credentials']:
                try:
                    credentials.append(JobGlobalCredentials.objects.get(id=creds_id))

                except ObjectDoesNotExist:
                    continue

            perm.credentials.set(credentials)

        if 'repositories' in validated_data and is_set(validated_data['repositories']):
            repositories = []
            for repo_id in validated_data['repositories']:
                try:
                    repositories.append(Repository.objects.get(id=repo_id))

                except ObjectDoesNotExist:
                    continue

            perm.repositories.set(repositories)

        if 'users' in validated_data and is_set(validated_data['users']):
            users = []
            for user_id in validated_data['users']:
                try:
                    users.append(USERS.objects.get(id=user_id))

                except ObjectDoesNotExist:
                    continue

            perm.users.set(users)

        if 'groups' in validated_data and is_set(validated_data['groups']):
            groups = []
            for group_id in validated_data['groups']:
                try:
                    groups.append(GROUPS.objects.get(id=group_id))

                except ObjectDoesNotExist:
                    continue

            perm.groups.set(groups)

        perm.save()


def build_permissions(perm_id_filter: int = None) -> (list, dict):
    permissions_raw = JobPermission.objects.all()
    permission_jobs_id = {permission.id: [] for permission in permissions_raw}
    permission_jobs_name = {permission.id: [] for permission in permissions_raw}
    permission_credentials_id = {permission.id: [] for permission in permissions_raw}
    permission_credentials_name = {permission.id: [] for permission in permissions_raw}
    permission_repositories_id = {permission.id: [] for permission in permissions_raw}
    permission_repositories_name = {permission.id: [] for permission in permissions_raw}
    permission_users_id = {permission.id: [] for permission in permissions_raw}
    permission_users_name = {permission.id: [] for permission in permissions_raw}
    permission_groups_id = {permission.id: [] for permission in permissions_raw}
    permission_groups_name = {permission.id: [] for permission in permissions_raw}

    for mapping in JobPermissionMapping.objects.all():
        permission_jobs_id[mapping.permission.id].append(mapping.job.id)
        permission_jobs_name[mapping.permission.id].append(mapping.job.name)

    for mapping in JobCredentialsPermissionMapping.objects.all():
        permission_credentials_id[mapping.permission.id].append(mapping.credentials.id)
        permission_credentials_name[mapping.permission.id].append(mapping.credentials.name)

    for mapping in JobRepositoryPermissionMapping.objects.all():
        permission_repositories_id[mapping.permission.id].append(mapping.repository.id)
        permission_repositories_name[mapping.permission.id].append(mapping.repository.name)

    for mapping in JobPermissionMemberUser.objects.all():
        permission_users_id[mapping.permission.id].append(mapping.user.id)
        permission_users_name[mapping.permission.id].append(mapping.user.username)

    for mapping in JobPermissionMemberGroup.objects.all():
        permission_groups_id[mapping.permission.id].append(mapping.group.id)
        permission_groups_name[mapping.permission.id].append(mapping.group.name)

    permissions = []

    for permission in permissions_raw:
        if perm_id_filter is not None:
            if perm_id_filter != permission.id:
                continue

        permissions.append({
            'id': permission.id,
            'name': permission.name,
            'permission': permission.permission,
            'permission_name': permission.permission_name,
            'jobs': permission_jobs_id[permission.id],
            'jobs_all': permission.jobs_all,
            'credentials': permission_credentials_id[permission.id],
            'credentials_all': permission.credentials_all,
            'repositories': permission_repositories_id[permission.id],
            'repositories_all': permission.repositories_all,
            'users': permission_users_id[permission.id],
            'groups': permission_groups_id[permission.id],
            'jobs_name': permission_jobs_name[permission.id],
            'credentials_name': permission_credentials_name[permission.id],
            'repositories_name': permission_repositories_name[permission.id],
            'users_name': permission_users_name[permission.id],
            'groups_name': permission_groups_name[permission.id],
        })

    try:
        if perm_id_filter is not None:
            return permissions[0]

    except IndexError:
        return {}

    return permissions


def permission_in_use(permission: JobPermission) -> bool:
    in_use_jobs = JobPermissionMapping.objects.filter(permission=permission).exists()
    in_use_creds = JobCredentialsPermissionMapping.objects.filter(permission=permission).exists()
    return in_use_jobs or in_use_creds


class APIPermission(GenericAPIView):
    http_method_names = ['get', 'post']
    serializer_class = PermissionReadResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={200: PermissionReadResponse},
        summary='Return list of permissions',
        operation_id='permission_list',
    )
    def get(request):
        del request
        return Response(build_permissions())

    @extend_schema(
        request=PermissionWriteRequest,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Permission created'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid permission data provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to create a permission'),
        },
        summary='Create a new Permission.',
        operation_id='permission_create',
    )
    def post(self, request):
        privileged = has_manager_privileges(user=get_api_user(request), kind='permission')
        if not privileged:
            return Response(
                data={'msg': 'Not privileged to manage permissions'},
                status=403,
            )

        serializer = PermissionWriteRequest(data=request.data)

        if not serializer.is_valid():
            return Response(
                data={'msg': f"Provided permission data is not valid: '{serializer.errors}'"},
                status=400,
            )

        try:
            serializer.create_or_update(validated_data=serializer.validated_data, perm=None)

        except IntegrityError as err:
            return Response(
                data={'msg': f"Provided permission data is not valid: '{err}'"},
                status=400,
            )

        return Response({'msg': f"Permission '{serializer.data['name']}' created successfully"})


class APIPermissionItem(GenericAPIView):
    http_method_names = ['get', 'put', 'delete']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={200: PermissionReadResponse},
        summary='Return information of a permission.',
        operation_id='permission_get'
    )
    def get(request, perm_id: int):
        del request
        return Response(build_permissions(perm_id_filter=perm_id))

    @extend_schema(
        request=PermissionWriteRequest,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Permission updated'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid permission data provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to edit the permission'),
            404: OpenApiResponse(response=GenericResponse, description='Permission does not exist'),
        },
        summary='Modify a permission.',
        operation_id='permission_edit',
    )
    def put(self, request, perm_id: int):
        privileged = has_manager_privileges(user=get_api_user(request), kind='permission')
        if not privileged:
            return Response(
                data={'msg': 'Not privileged to manage permissions'},
                status=403,
            )

        serializer = PermissionWriteRequest(data=request.data)

        if not serializer.is_valid():
            return Response(
                data={'msg': f"Provided permission data is not valid: '{serializer.errors}'"},
                status=400,
            )

        try:
            permission = JobPermission.objects.get(id=perm_id)

        except ObjectDoesNotExist:
            permission = None

        if permission is None:
            return Response(
                data={'msg': f"Permission with ID {perm_id} does not exist"},
                status=404,
            )

        try:
            serializer.create_or_update(validated_data=serializer.validated_data, perm=permission)
            return Response(data={'msg': f"Permission '{permission.name}' updated"}, status=200)

        except IntegrityError as err:
            return Response(data={'msg': f"Provided permission data is not valid: '{err}'"}, status=400)

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Permission deleted'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid permission data provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to delete the permission'),
            404: OpenApiResponse(response=GenericResponse, description='Permission does not exist'),
        },
        summary='Delete a permission.',
        operation_id='permission_delete'
    )
    def delete(self, request, perm_id: int):
        privileged = has_manager_privileges(user=get_api_user(request), kind='permission')
        if not privileged:
            return Response(
                data={'msg': 'Not privileged to manage permissions'},
                status=403,
            )

        try:
            permission = JobPermission.objects.get(id=perm_id)
            if permission is not None:
                if permission_in_use(permission):
                    return Response(
                        data={'msg': f"Permission '{permission.name}' cannot be deleted as it is still in use"},
                        status=400,
                    )

                permission.delete()
                return Response(data={'msg': f"Permission '{permission.name}' deleted"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Permission with ID {perm_id} does not exist"}, status=404)



import signal
from os import environ
from os import kill as os_kill
from sys import exit as sys_exit
from threading import Thread, ThreadError
from time import sleep, time

from gunicorn.arbiter import Arbiter
from django.core.validators import ValidationError
from django.db.utils import OperationalError, IntegrityError

from aw.settings import DB_FILE
from aw.execute.threader import ThreadManager
from aw.utils.debug import log
from aw.utils.util import is_null
from aw.config.hardcoded import INTERVAL_CHECK, INTERVAL_RELOAD
from aw.model.job import Job, JobExecution, validate_cronjob
from aw.execute.queue import queue_get


class Scheduler:
    WAIT_TIME = 1

    def __init__(self):
        self.thread_manager = ThreadManager()
        self.stopping = False
        self.reloading = False

    def stop(self, signum=None, error: bool = False):
        if not self.stopping:
            log('Stopping..', level=3)
            self._signum_log(signum=signum)
            self.stopping = True

            log('Stopping job-threads', level=6)
            self.thread_manager.stop()

            sleep(self.WAIT_TIME)
            log('Finished!')

            # else gunicorn will not know it should die
            os_kill(int(environ['MAINPID']), signal.SIGTERM)

            # just in case
            if error or signum is not None:
                sys_exit(1)

            sys_exit(0)

    @staticmethod
    def _signum_log(signum):
        log(f'Scheduler got signal {signum}')

    def _add_thread(self, job: Job, execution: JobExecution = None, once: bool = False):
        self.thread_manager.add_thread(job=job, execution=execution, once=once)
        self.thread_manager.start_thread(job=job)

    def start(self):
        log('Starting..', level=3)
        log('Starting job-threads', level=4)
        try:
            self.reload()
            self._run()

        except (OperationalError, IntegrityError) as err:
            log(
                "Database has an unexpected state! "
                f"If you are fine with losing the existing config - delete the database file: {DB_FILE}\n"
                f"Error: {err}"
            )
            self.stop()

    def _run(self):
        # pylint: disable=W0718
        try:
            sleep(self.WAIT_TIME)
            log('Entering Scheduler runtime', level=7)
            time_last_check = time()
            time_last_reload = time()

            while True:
                try:
                    if self.stopping:
                        break

                    if time() > (time_last_check + INTERVAL_CHECK):
                        self.check()
                        time_last_check = time()

                    if time() > (time_last_reload + INTERVAL_RELOAD):
                        self.reload()
                        time_last_reload = time()

                    sleep(self.WAIT_TIME)

                except ThreadError as err:
                    log(msg=f'Got thread error: {err}', level=2)

        except Exception as err:
            log(msg=f'Got unexpected error: {err}', level=1)
            self.stop(error=True)
            return

    def status(self):
        log(msg=f"Running job-threads: {self.thread_manager.list_pretty()}", level=4)

    def check(self):
        log('Checking for queued jobs', level=7)
        while True:
            queue_item = queue_get()
            if queue_item is None:
                break

            job, user = queue_item

            log(f"Adding job-thread for queued job: '{job.name}' (triggered by user '{user.username}')", level=4)
            self._add_thread(job=job, execution=JobExecution(user=user, job=job, comment='Triggered'), once=True)

    def reload(self, signum=None):
        if not self.reloading and not self.stopping:
            self.reloading = True

            if signum is not None:
                log('Reloading..', level=3)
                self._signum_log(signum=signum)

            self._reload_action(**self._reload_check())
            self.thread_manager.clean_stopped_threads()
            self.reloading = False

    def _reload_action(self, added: list, removed: list, changed: list):
        any_changed = False

        log('Checking jobs for config-changes', level=7)
        if len(added) > 0:
            any_changed = True
            log(f"Adding job-threads: {[job.name for job in added]}", level=4)
            for job in added:
                self._add_thread(job=job)

        if len(removed) > 0:
            any_changed = True
            log(f"Removing job-threads: {[job.name for job in removed]}", level=4)
            for job in removed:
                self.thread_manager.stop_thread(job)

        if len(changed) > 0:
            any_changed = True
            log(f"Replacing job-threads: {[job.name for job in changed]}", level=4)
            for job in changed:
                self.thread_manager.replace_thread(job)

        if any_changed:
            sleep(1)  # wait for threads to initialize
            self.status()

    def _reload_check(self) -> dict:
        result = {'added': [], 'removed': [], 'changed': []}
        running = self.thread_manager.list()
        running_ids = [job.id for job in running]
        configured = Job.objects.all()
        configured_ids = [job.id for job in configured]

        for job in configured:
            if job.id not in running_ids:
                if is_null(job.schedule):
                    log(f"Ignoring job '{job.name}' because it has no schedule", level=6)
                    continue

                if not job.enabled:
                    log(f"Ignoring job '{job.name}' because it is disabled", level=6)
                    continue

                try:
                    validate_cronjob(job.schedule)
                    result['added'].append(job)

                except ValidationError:
                    log(f"Got invalid job schedule '{job.schedule}'", level=4)

            else:
                run_job = [run_job for run_job in running if run_job.id == job.id][0]
                for field in Job.CHANGE_FIELDS:
                    if getattr(run_job, field) != getattr(job, field):
                        if run_job.enabled and not job.enabled:
                            result['removed'].append(job)
                            log(f"Job '{job.name}' was disabled", level=6)
                            break

                        result['changed'].append(job)
                        log(f"Job '{job.name}' config changed", level=6)
                        break

        for job in running:
            if job.id not in configured_ids:
                result['removed'].append(job)

        return result


def init_scheduler():
    scheduler = Scheduler()
    scheduler_thread = Thread(target=scheduler.start)

    # override gunicorn signal handling to allow for graceful shutdown
    Arbiter.SIGNALS.remove(signal.SIGHUP)
    Arbiter.SIGNALS.remove(signal.SIGINT)
    Arbiter.SIGNALS.remove(signal.SIGTERM)

    def signal_exit(signum=None, stack=None):
        del stack
        scheduler.stop(signum)
        os_kill(int(environ['MAINPID']), signal.SIGQUIT)  # trigger 'Arbiter.stop'
        sleep(3)
        sys_exit(1)

    def signal_reload(signum=None, stack=None):
        del stack
        scheduler.reload(signum)

    signal.signal(signal.SIGHUP, signal_reload)
    signal.signal(signal.SIGINT, signal_exit)
    signal.signal(signal.SIGTERM, signal_exit)

    scheduler_thread.start()
